import json
import os
import random
from pathlib import Path

from Class import settingkey
from Class.openkhmod import Asset
from Class.seedSettings import SeedSettings
from Module import appconfig
from Module.cosmeticsmods import music
from Module.cosmeticsmods.endingpic import EndingPictureRandomizer
from Module.cosmeticsmods.field2d import CommandMenuRandomizer, RoomTransitionImageRandomizer
from Module.cosmeticsmods.itempic import ItempicRandomizer
from Module.cosmeticsmods.keyblade import KeybladeRandomizer


class CustomCosmetics:

    def __init__(self):
        super().__init__()

        config_path = Path('auto-save') / 'custom-cosmetics.json'
        self.config_path = config_path
        self.external_executables: list[str] = []

        if config_path.is_file():
            with open(config_path, encoding='utf-8') as config_file:
                try:
                    raw_json: dict = json.load(config_file)
                    self.external_executables = raw_json.get('external_executables', [])
                except Exception:
                    print('Error decoding custom-cosmetics.json, using empty list of external executables')

    def add_custom_executable(self, path_str: str):
        self.external_executables.append(path_str)

    def remove_executable_at_index(self, index: int):
        del self.external_executables[index]

    def write_file(self):
        with open(self.config_path, 'w', encoding='utf-8') as config_file:
            raw_json = {
                'external_executables': self.external_executables
            }
            json.dump(raw_json, config_file, indent=4)

    def collect_custom_executable_files(self) -> list[str]:
        return [custom_file for custom_file in self.external_executables]


class CosmeticsMod:

    @staticmethod
    def bootstrap_cosmetics_files():
        CosmeticsMod.bootstrap_music_list_file()
        ItempicRandomizer.bootstrap_itempic_file()

        custom_music_path = appconfig.read_custom_music_path()
        if custom_music_path is not None:
            CosmeticsMod.bootstrap_custom_music_folder(custom_music_path)

        custom_visuals_path = appconfig.read_custom_visuals_path()
        if custom_visuals_path is not None:
            CosmeticsMod.bootstrap_custom_visuals_folder(custom_visuals_path)

    @staticmethod
    def bootstrap_music_list_file() -> Path:
        """Creates the musiclist file if needed."""
        music_list_file_path = Path('musiclist.json')
        if not music_list_file_path.is_file():
            with open(music_list_file_path, mode='w', encoding='utf-8') as music_list_file:
                json.dump(music.kh2_music_list, music_list_file, indent=4)
        return music_list_file_path

    @staticmethod
    def bootstrap_custom_music_folder(custom_music_path: Path):
        """Creates folders for each of the default music categories."""
        for folder in ['atlantica', 'battle', 'boss', 'cutscene', 'field', 'title', 'wild']:
            (custom_music_path / folder).mkdir(exist_ok=True)

    @staticmethod
    def randomize_music(ui_settings: SeedSettings) -> tuple[list[Asset], dict[str, str]]:
        """
        Randomizes music, returning a list of assets to be added to the seed mod and a dictionary of which song was
        replaced by which replacement.
        """
        return CosmeticsMod._get_music_assets(ui_settings)

    @staticmethod
    def get_keyblade_summary() -> dict[str, int]:
        vanilla = KeybladeRandomizer.collect_vanilla_keyblades()
        custom = KeybladeRandomizer.collect_custom_keyblades()
        return {"In-Game": len(vanilla), "Custom": len(custom)}

    @staticmethod
    def get_music_summary(settings: SeedSettings) -> dict[str, int]:
        music_files = CosmeticsMod._collect_music_files(settings)

        return {category: len(song_list) for category, song_list in music_files.items()}

    @staticmethod
    def _collect_music_files(settings: SeedSettings) -> dict[str, list[Path]]:
        """Returns music files grouped by category."""
        result: dict[str, list[Path]] = {}

        categorize = settings.get(settingkey.MUSIC_RANDO_PC_USE_CATEGORIES)
        dmca_safe = settings.get(settingkey.MUSIC_RANDO_PC_DMCA_SAFE)

        if settings.get(settingkey.MUSIC_RANDO_PC_INCLUDE_CUSTOM):
            custom_music_path = appconfig.read_custom_music_path()
            if custom_music_path is not None:
                for child in [str(category_file).lower() for category_file in os.listdir(custom_music_path)]:
                    child_path = custom_music_path / child
                    if child_path.is_dir():
                        category_songs: list[Path] = []
                        for root, dirs, files in os.walk(child_path):
                            root_path = Path(root)
                            for file in files:
                                _, extension = os.path.splitext(file)
                                if extension.lower() == '.scd':
                                    file_path = root_path / file
                                    category_songs.append(file_path)
                        resolved_category = child
                        if not categorize:
                            resolved_category = 'wild'
                        if resolved_category not in result:
                            result[resolved_category] = []
                        result[resolved_category] += category_songs

        extracted_data_path = appconfig.extracted_data_path()
        if extracted_data_path is not None:
            def add_game_song(song_file_path: Path, category: str, song_dmca: bool):
                if not song_file_path.is_file():
                    return

                if not categorize:
                    category = 'wild'
                if category not in result:
                    result[category] = []
                if not dmca_safe or not song_dmca:
                    result[category].append(song_file_path)

            if settings.get(settingkey.MUSIC_RANDO_PC_INCLUDE_KH2):
                kh2_path = extracted_data_path / 'kh2'
                if kh2_path.is_dir():
                    for kh2_song in music.kh2_music_list:
                        add_game_song(
                            song_file_path=kh2_path / kh2_song['filename'],
                            category=kh2_song['type'][0].lower(),
                            song_dmca=kh2_song.get('dmca', False)
                        )

            def add_other_game_music(enabled_key: str, game_music_path: Path, game_music_list: list):
                if settings.get(enabled_key) and game_music_path.is_dir():
                    for song in game_music_list:
                        add_game_song(
                            song_file_path=game_music_path / song['name'],
                            category=song['kind'],
                            song_dmca=song.get('dmca', False)
                        )

            add_other_game_music(
                enabled_key=settingkey.MUSIC_RANDO_PC_INCLUDE_KH1,
                game_music_path=extracted_data_path / 'kh1' / 'remastered' / 'amusic',
                game_music_list=music.kh1_music_list
            )
            add_other_game_music(
                enabled_key=settingkey.MUSIC_RANDO_PC_INCLUDE_RECOM,
                game_music_path=extracted_data_path / 'recom' / 'STREAM' / '0001',
                game_music_list=music.recom_music_list
            )
            add_other_game_music(
                enabled_key=settingkey.MUSIC_RANDO_PC_INCLUDE_BBS,
                game_music_path=extracted_data_path / 'bbs' / 'sound' / 'win' / 'bgm',
                game_music_list=music.bbs_music_list
            )
            add_other_game_music(
                enabled_key=settingkey.MUSIC_RANDO_PC_INCLUDE_DDD,
                game_music_path=extracted_data_path / 'kh3d' / 'sound' / 'jp' / 'output' / 'BGM',
                game_music_list=music.ddd_music_list
            )

        return result

    @staticmethod
    def _get_music_assets(settings: SeedSettings) -> tuple[list[Asset], dict[str, str]]:
        music_rando_enabled = settings.get(settingkey.MUSIC_RANDO_ENABLED_PC)
        if not music_rando_enabled:
            return [], {}

        allow_duplicates = settings.get(settingkey.MUSIC_RANDO_PC_ALLOW_DUPLICATES)

        # Primary copy of the music. Songs are removed as they are chosen.
        music_files_by_categories: dict[str, list[Path]] = {}
        # Secondary copy of the music. Used to refill the lists once all are gone.
        backup_files_by_categories: dict[str, list[Path]] = {}

        music_files = CosmeticsMod._collect_music_files(settings)
        for category, song_list in music_files.items():
            main_list = song_list.copy()
            random.shuffle(main_list)
            music_files_by_categories[category] = main_list

            backup_files_by_categories[category] = song_list

        assets: list[Asset] = []
        replacements: dict[str, str] = {}

        music_list_file_path = CosmeticsMod.bootstrap_music_list_file()
        with open(music_list_file_path, encoding='utf-8') as music_list_file:
            music_metadata = json.load(music_list_file)
        random.shuffle(music_metadata)
        for info in music_metadata:
            filename = info['filename']
            title = info['title']
            types = [song_type.lower() for song_type in info['type'] if song_type.lower() in music_files_by_categories]
            if len(types) > 0:
                random.shuffle(types)

                for chosen_type in types:
                    songs_for_chosen_type = music_files_by_categories[chosen_type]

                    # If there aren't any songs left for the type, attempt to refill the list if duplicates are allowed.
                    # If duplicates are not allowed and we run out of replacements, more will end up un-randomized.
                    if len(songs_for_chosen_type) == 0 and allow_duplicates:
                        refill_list = backup_files_by_categories[chosen_type].copy()
                        random.shuffle(refill_list)
                        music_files_by_categories[chosen_type] = refill_list
                        songs_for_chosen_type = refill_list

                    if len(songs_for_chosen_type) > 0:
                        chosen_song = str(songs_for_chosen_type.pop())

                        asset = {
                            'name': filename,
                            'vanilla_song': title,
                            'song_type': chosen_type,
                            'method': 'copy',
                            'source': [{'name': chosen_song}]
                        }
                        assets.append(asset)

                        replacements[title] = chosen_song

                        break

        return assets, replacements

    @staticmethod
    def bootstrap_custom_visuals_folder(custom_visuals_path: Path):
        """Creates folders for each of the default custom visuals categories."""
        for folder in ["room-transition-images"]:
            (custom_visuals_path / folder).mkdir(exist_ok=True)

        end_pictures_path = custom_visuals_path / "ending-pictures"
        end_pictures_path.mkdir(exist_ok=True)

        item_pictures_path = custom_visuals_path / "item-pictures"
        item_pictures_path.mkdir(exist_ok=True)

        itempic_categories: list[str] = []
        for itempic in ItempicRandomizer.replaceable_itempics():
            itempic_categories.extend(itempic.types)
        for category in set(itempic_categories):
            (item_pictures_path / category.lower()).mkdir(exist_ok=True)

    @staticmethod
    def randomize_keyblades(seed_settings: SeedSettings) -> tuple[list[Asset], dict[str, str]]:
        """
        Randomizes keyblades, returning a list of assets to be added to the seed mod and a dictionary of which keyblade
        was replaced by which replacement.
        """
        return KeybladeRandomizer.randomize_keyblades(
            setting=seed_settings.get(settingkey.KEYBLADE_RANDO),
            include_effects=seed_settings.get(settingkey.KEYBLADE_RANDO_INCLUDE_EFFECTS),
            allow_duplicate_replacement=seed_settings.get(settingkey.KEYBLADE_RANDO_ALLOW_DUPLICATES)
        )

    @staticmethod
    def randomize_field2d(seed_settings: SeedSettings) -> list[Asset]:
        """Randomizes various field2d entries, returning a list of assets to be added to a mod."""
        assets: list[Asset] = []

        command_menu_choice = seed_settings.get(settingkey.COMMAND_MENU)
        assets.extend(CommandMenuRandomizer(command_menu_choice).randomize_command_menus())

        transition_choice = seed_settings.get(settingkey.ROOM_TRANSITION_IMAGES)
        assets.extend(RoomTransitionImageRandomizer(transition_choice).randomize_room_transitions())

        return assets

    @staticmethod
    def randomize_itempics(seed_settings: SeedSettings) -> list[Asset]:
        """Randomizes various itempic entries, returning a list of assets to be added to a mod."""
        setting = seed_settings.get(settingkey.ITEMPIC_RANDO)
        return ItempicRandomizer.randomize_itempics(setting)

    @staticmethod
    def randomize_end_screen(seed_settings: SeedSettings) -> list[Asset]:
        """Randomizes the ending screen, returning a list of assets to be added to a mod."""
        setting = seed_settings.get(settingkey.ENDPIC_RANDO)
        return EndingPictureRandomizer.randomize_end_screen(setting)



from enum import Enum
from typing import Optional

from Class.newLocationClass import KH2Location
from List.configDict import locationType
from List.inventory import ability, form as driveform, growth
from List.inventory.form import DriveForm
from List.inventory.item import InventoryItem
from List.location.graph import RequirementEdge, LocationGraphBuilder, START_NODE
from Module.itemPlacementRestriction import ItemPlacementHelpers


class CheckLocation(str, Enum):
    Valor2 = "Valor Level 2"
    Valor3 = "Valor Level 3"
    Valor4 = "Valor Level 4"
    Valor5 = "Valor Level 5"
    Valor6 = "Valor Level 6"
    Valor7 = "Valor Level 7"
    Wisdom2 = "Wisdom Level 2"
    Wisdom3 = "Wisdom Level 3"
    Wisdom4 = "Wisdom Level 4"
    Wisdom5 = "Wisdom Level 5"
    Wisdom6 = "Wisdom Level 6"
    Wisdom7 = "Wisdom Level 7"
    Limit2 = "Limit Level 2"
    Limit3 = "Limit Level 3"
    Limit4 = "Limit Level 4"
    Limit5 = "Limit Level 5"
    Limit6 = "Limit Level 6"
    Limit7 = "Limit Level 7"
    Master2 = "Master Level 2"
    Master3 = "Master Level 3"
    Master4 = "Master Level 4"
    Master5 = "Master Level 5"
    Master6 = "Master Level 6"
    Master7 = "Master Level 7"
    Final2 = "Final Level 2"
    Final3 = "Final Level 3"
    Final4 = "Final Level 4"
    Final5 = "Final Level 5"
    Final6 = "Final Level 6"
    Final7 = "Final Level 7"


def form_level(level: int, form: DriveForm, vanilla: Optional[InventoryItem] = None) -> KH2Location:
    types = [locationType.FormLevel]
    if level == 1:
        types.append(locationType.FormLevel1)

    vanilla_items: list[int] = []
    if vanilla is not None:
        vanilla_items.append(vanilla.id)

    description = f"{form.short_name} Level {level}"
    return KH2Location(level, description, form.level_category, types, VanillaItems=vanilla_items)


def make_graph(graph: LocationGraphBuilder):
    forms_with_vanilla_rewards = {
        driveform.ValorForm: {
            1: None,
            2: ability.AutoValor,
            3: growth.HighJump2,
            4: ability.ComboPlus,
            5: growth.HighJump3,
            6: ability.ComboPlus,
            7: growth.HighJumpMax
        },
        driveform.WisdomForm: {
            1: None,
            2: ability.AutoWisdom,
            3: growth.QuickRun2,
            4: ability.MpRage,
            5: growth.QuickRun3,
            6: ability.MpHaste,
            7: growth.QuickRunMax
        },
        driveform.LimitForm: {
            1: None,
            2: ability.AutoLimitForm,
            3: growth.DodgeRoll2,
            4: ability.Draw,
            5: growth.DodgeRoll3,
            6: ability.LuckyLucky,
            7: growth.DodgeRollMax
        },
        driveform.MasterForm: {
            1: None,
            2: ability.AutoMaster,
            3: growth.AerialDodge2,
            4: ability.AirComboPlus,
            5: growth.AerialDodge3,
            6: ability.AirComboPlus,
            7: growth.AerialDodgeMax
        },
        driveform.FinalForm: {
            1: None,
            2: ability.AutoFinal,
            3: growth.Glide2,
            4: ability.FormBoost,
            5: growth.Glide3,
            6: ability.FormBoost,
            7: growth.GlideMax
        }
    }

    form_helper = ItemPlacementHelpers.make_form_lambda
    settings = graph.settings
    extended_placement_logic = settings.extended_placement_logic
    if extended_placement_logic and settings.disable_final_form:
        form_helper = ItemPlacementHelpers.make_form_lambda_nightmare_no_final
    elif extended_placement_logic and settings.disable_antiform:
        form_helper = ItemPlacementHelpers.make_form_lambda_nightmare_no_anti
    elif extended_placement_logic:
        form_helper = ItemPlacementHelpers.make_form_lambda_nightmare

    for form, vanilla_rewards in forms_with_vanilla_rewards.items():
        locations: dict[int, str] = {}
        for level in range(1, 8):
            locations[level] = graph.add_location(f"{form.short_name}-{level}", [
                form_level(level, form, vanilla=vanilla_rewards[level])
            ])

        graph.add_edge(START_NODE, locations[1])
        for level in range(2, 8):
            graph.add_edge(locations[level - 1], locations[level], RequirementEdge(req=form_helper(form, level)))



from dataclasses import dataclass, field

from List.configDict import itemType
from List.inventory.item import InventoryItem


@dataclass(frozen=True)
class Accessory(InventoryItem):
    id: int
    name: str
    type: itemType = field(init=False, default=itemType.ACCESSORY)


AbilityRing = Accessory(8, "Ability Ring")
EngineersRing = Accessory(9, "Engineer's Ring")
TechniciansRing = Accessory(10, "Technician's Ring")
ExpertsRing = Accessory(11, "Expert's Ring")
SardonyxRing = Accessory(12, "Sardonyx Ring")
TourmalineRing = Accessory(13, "Tourmaline Ring")
AquamarineRing = Accessory(14, "Aquamarine Ring")
GarnetRing = Accessory(15, "Garnet Ring")
DiamondRing = Accessory(16, "Diamond Ring")
SilverRing = Accessory(17, "Silver Ring")
GoldRing = Accessory(18, "Gold Ring")
PlatinumRing = Accessory(19, "Platinum Ring")
MythrilRing = Accessory(20, "Mythril Ring")
OrichalcumRing = Accessory(28, "Orichalcum Ring")
MastersRing = Accessory(34, "Master's Ring")
MoonAmulet = Accessory(35, "Moon Amulet")
StarCharm = Accessory(36, "Star Charm")
SkillRing = Accessory(38, "Skill Ring")
SkillfulRing = Accessory(39, "Skillful Ring")
SoldierEarring = Accessory(40, "Soldier Earring")
FencerEarring = Accessory(46, "Fencer Earring")
MageEarring = Accessory(47, "Mage Earring")
SlayerEarring = Accessory(48, "Slayer Earring")
CosmicRing = Accessory(52, "Cosmic Ring")
Medal = Accessory(53, "Medal")
CosmicArts = Accessory(56, "Cosmic Arts")
ShadowArchive = Accessory(57, "Shadow Archive")
ShadowArchivePlus = Accessory(58, "Shadow Archive+")
LuckyRing = Accessory(63, "Lucky Ring")
FullBloom = Accessory(64, "Full Bloom")
DrawRing = Accessory(65, "Draw Ring")
FullBloomPlus = Accessory(66, "Full Bloom+")
ExecutivesRing = Accessory(599, "Executive's Ring")



from enum import Enum

from Class.newLocationClass import KH2Location
from List.configDict import locationType
from List.inventory import proof
from List.location.graph import chest, LocationGraphBuilder, START_NODE


class CheckLocation(str, Enum):
    GardenOfAssemblageMap = "Garden of Assemblage Map"
    GoaLostIllusion = "GoA Lost Illusion"
    ProofOfNonexistence = "Proof of Nonexistence"


def make_graph(graph: LocationGraphBuilder):
    locations = [chest(i, f"Starting Item {i}", locationType.Critical) for i in range(1, 8)]
    locations.append(chest(585, CheckLocation.GardenOfAssemblageMap, locationType.Free))
    locations.append(chest(586, CheckLocation.GoaLostIllusion, locationType.Free))
    locations.append(
        chest(590, CheckLocation.ProofOfNonexistence, locationType.Free, vanilla=proof.ProofOfNonexistence))

    graph.add_location(START_NODE, locations)


def donald_starting_items() -> list[KH2Location]:
    return [chest(i, f"Donald Starting Item {i}", locationType.Free) for i in range(1, 3)]


def goofy_starting_items() -> list[KH2Location]:
    return [chest(i, f"Goofy Starting Item {i}", locationType.Free) for i in range(1, 3)]



from dataclasses import dataclass, field

from List.configDict import itemType
from List.inventory.item import InventoryItem


@dataclass(frozen=True)
class SummonCharm(InventoryItem):
    id: int
    name: str
    type: itemType = field(init=False, default=itemType.SUMMON)


UkuleleCharm = SummonCharm(25, "Ukulele Charm (Stitch)")
LampCharm = SummonCharm(159, "Lamp Charm (Genie)")
FeatherCharm = SummonCharm(160, "Feather Charm (Peter Pan)")
BaseballCharm = SummonCharm(383, "Baseball Charm (Chicken Little)")


def all_summon_charms() -> list[SummonCharm]:
    return [LampCharm, FeatherCharm, UkuleleCharm, BaseballCharm]



from enum import Enum

from List.configDict import locationType
from List.inventory import magic, keyblade, ability, report
from List.location.graph import DefaultLogicGraph, RequirementEdge, chest, popup, item_bonus, stat_bonus, hybrid_bonus, \
    LocationGraphBuilder, START_NODE
from Module.itemPlacementRestriction import ItemPlacementHelpers


class NodeId(str, Enum):
    BeastsCastleCourtyard = "BC Courtyard"
    BeastsCastleCourtyardChests = "BC Courtyard Chests"
    BellesRoom = "Belle's Room"
    BellesRoomChests = "Belle's Room Chests"
    EastWing = "East Wing"
    EastWingChests = "East Wing Chests"
    WestHall = "West Hall"
    WestHallChests = "West Hall Chests"
    Thresholder = "Thresholder"
    Dungeon = "Dungeon"
    DungeonChests = "Dungeon Chests"
    SecretPassage = "Secret Passage"
    SecretPassageChests = "Secret Passage Chests"
    WestHallPostDungeon = "West Hall Post Dungeon"
    WestHallPostDungeonChests = "West Hall Post Dungeon Chests"
    WestWing = "West Wing"
    WestWingChests = "West Wing Chests"
    BeastBonus = "Beast Bonus"
    BeastsRoom = "Beast's Room"
    BeastsRoomChests = "Beast's Room Chests"
    DarkThorn = "Dark Thorn"
    RumblingRose = "Rumbling Rose"
    Xaldin = "Xaldin"
    DataXaldin = "Data Xaldin"


class CheckLocation(str, Enum):
    CourtyardApBoost = "BC Courtyard AP Boost"
    CourtyardHiPotion = "BC Courtyard Hi-Potion"
    CourtyardMythrilShard = "BC Courtyard Mythril Shard"
    BellesRoomCastleMap = "Belle's Room Castle Map"
    BellesRoomMegaRecipe = "Belle's Room Mega-Recipe"
    EastWingMythrilShard = "The East Wing Mythril Shard"
    EastWingTent = "The East Wing Tent"
    WestHallHiPotion = "The West Hall Hi-Potion"
    WestHallPowerShard = "The West Hall Power Shard"
    WestHallMythrilShard2 = "The West Hall Mythril Shard (2)"
    WestHallBrightStone = "The West Hall Bright Stone"
    WestHallMythrilShard = "The West Hall Mythril Shard"
    Thresholder = "Thresholder"
    DungeonBasementMap = "Dungeon Basement Map"
    DungeonApBoost = "Dungeon AP Boost"
    SecretPassageMythrilShard = "Secret Passage Mythril Shard"
    SecretPassageHiPotion = "Secret Passage Hi-Potion"
    SecretPassageLucidShard = "Secret Passage Lucid Shard"
    WestHallPostDungeonApBoost = "The West Hall AP Boost (Post Dungeon)"
    WestWingMythrilShard = "The West Wing Mythril Shard"
    WestWingTent = "The West Wing Tent"
    Beast = "Beast"
    BeastsRoomBlazingShard = "The Beast's Room Blazing Shard"
    DarkThorn = "Dark Thorn Bonus"
    DarkThornCureElement = "Dark Thorn Cure Element"
    RumblingRose = "Rumbling Rose"
    CastleWallsMap = "Castle Walls Map"
    XaldinBonus = "Xaldin Bonus"
    XaldinSecretAnsemReport4 = "Secret Ansem Report 4"
    DataXaldin = "Xaldin (Data) Defense Boost"

class BCLogicGraph(DefaultLogicGraph):
    def __init__(self,reverse_rando,keyblade_unlocks):
        DefaultLogicGraph.__init__(self,NodeId)
        keyblade_lambda = lambda inv : not keyblade_unlocks or ItemPlacementHelpers.need_bc_keyblade(inv)
        self.logic[NodeId.BeastsCastleCourtyard][NodeId.BeastsCastleCourtyardChests] = keyblade_lambda
        self.logic[NodeId.BellesRoom][NodeId.BellesRoomChests] = keyblade_lambda
        self.logic[NodeId.EastWing][NodeId.EastWingChests] = keyblade_lambda
        self.logic[NodeId.WestHall][NodeId.WestHallChests] = keyblade_lambda
        self.logic[NodeId.Dungeon][NodeId.DungeonChests] = keyblade_lambda
        self.logic[NodeId.SecretPassage][NodeId.SecretPassageChests] = keyblade_lambda
        self.logic[NodeId.WestHallPostDungeon][NodeId.WestHallPostDungeonChests] = keyblade_lambda
        self.logic[NodeId.WestWing][NodeId.WestWingChests] = keyblade_lambda
        self.logic[NodeId.BeastsRoom][NodeId.BeastsRoomChests] = keyblade_lambda
        if not reverse_rando:
            self.logic[START_NODE][NodeId.BeastsCastleCourtyard] = ItemPlacementHelpers.bc1_check
            self.logic[NodeId.DarkThorn][NodeId.RumblingRose] = ItemPlacementHelpers.bc2_check
        else:
            self.logic[START_NODE][NodeId.WestHall] = ItemPlacementHelpers.bc1_check
            self.logic[NodeId.Xaldin][NodeId.BeastsCastleCourtyard] = ItemPlacementHelpers.bc2_check

def make_graph(graph: LocationGraphBuilder):
    bc = locationType.BC
    bc_logic = BCLogicGraph(graph.reverse_rando,graph.keyblades_unlock_chests)
    graph.add_logic(bc_logic)

    courtyard_chests = graph.add_location(NodeId.BeastsCastleCourtyardChests, [
        chest(39, CheckLocation.CourtyardApBoost, bc),
        chest(40, CheckLocation.CourtyardHiPotion, bc),
        chest(505, CheckLocation.CourtyardMythrilShard, bc),
    ])
    courtyard = graph.add_location(NodeId.BeastsCastleCourtyard, [])
    belles_room_chests = graph.add_location(NodeId.BellesRoomChests, [
        chest(46, CheckLocation.BellesRoomCastleMap, bc),
        chest(240, CheckLocation.BellesRoomMegaRecipe, bc),
    ])
    belles_room = graph.add_location(NodeId.BellesRoom, [])
    east_wing_chests = graph.add_location(NodeId.EastWingChests, [
        chest(63, CheckLocation.EastWingMythrilShard, bc),
        chest(155, CheckLocation.EastWingTent, bc),
    ])
    east_wing = graph.add_location(NodeId.EastWing, [])
    west_hall_chests = graph.add_location(NodeId.WestHallChests, [
        chest(41, CheckLocation.WestHallHiPotion, bc),
        chest(207, CheckLocation.WestHallPowerShard, bc),
        chest(208, CheckLocation.WestHallMythrilShard2, bc),
        chest(159, CheckLocation.WestHallBrightStone, bc),
        chest(206, CheckLocation.WestHallMythrilShard, bc),
    ])
    west_hall = graph.add_location(NodeId.WestHall, [])
    thresholder = graph.add_location(NodeId.Thresholder, [
        item_bonus(2, CheckLocation.Thresholder, bc, vanilla=ability.UpperSlash),
    ])
    dungeon_chests = graph.add_location(NodeId.DungeonChests, [
        chest(239, CheckLocation.DungeonBasementMap, bc),
        chest(43, CheckLocation.DungeonApBoost, bc),
    ])
    dungeon = graph.add_location(NodeId.Dungeon, [])
    secret_passage_chests = graph.add_location(NodeId.SecretPassageChests, [
        chest(44, CheckLocation.SecretPassageMythrilShard, bc),
        chest(168, CheckLocation.SecretPassageHiPotion, bc),
        chest(45, CheckLocation.SecretPassageLucidShard, bc),
    ])
    secret_passage = graph.add_location(NodeId.SecretPassage, [])
    west_hall_post_dungeon_chests = graph.add_location(NodeId.WestHallPostDungeonChests, [
        chest(158, CheckLocation.WestHallPostDungeonApBoost, bc),
    ])
    west_hall_post_dungeon = graph.add_location(NodeId.WestHallPostDungeon, [])
    west_wing_chests = graph.add_location(NodeId.WestWingChests, [
        chest(42, CheckLocation.WestWingMythrilShard, bc),
        chest(164, CheckLocation.WestWingTent, bc),
    ])
    west_wing = graph.add_location(NodeId.WestWing, [])
    beast_bonus = graph.add_location(NodeId.BeastBonus, [
        stat_bonus(12, CheckLocation.Beast, bc),
    ])
    beasts_room_chests = graph.add_location(NodeId.BeastsRoomChests, [
        chest(241, CheckLocation.BeastsRoomBlazingShard, bc),
    ])
    beasts_room = graph.add_location(NodeId.BeastsRoom, [])
    dark_thorn = graph.add_location(NodeId.DarkThorn, [
        hybrid_bonus(3, CheckLocation.DarkThorn, bc, vanilla=ability.RetaliatingSlash),
        popup(299, CheckLocation.DarkThornCureElement, bc, vanilla=magic.Cure),
    ])
    rumbling_rose = graph.add_location(NodeId.RumblingRose, [
        popup(270, CheckLocation.RumblingRose, bc, vanilla=keyblade.RumblingRose),
        popup(325, CheckLocation.CastleWallsMap, bc),
    ])
    xaldin = graph.add_location(NodeId.Xaldin, [
        hybrid_bonus(4, CheckLocation.XaldinBonus, bc, vanilla=magic.Reflect),
        popup(528, CheckLocation.XaldinSecretAnsemReport4, bc, vanilla=report.AnsemReport4),
    ])
    data_xaldin = graph.add_location(NodeId.DataXaldin, [
        popup(559, CheckLocation.DataXaldin, [bc, locationType.DataOrg]),
    ])

    graph.register_superboss(data_xaldin)
    
    graph.add_edge(courtyard, courtyard_chests)
    graph.add_edge(belles_room, belles_room_chests)
    graph.add_edge(east_wing, east_wing_chests)
    graph.add_edge(west_hall, west_hall_chests)
    graph.add_edge(dungeon, dungeon_chests)
    graph.add_edge(secret_passage, secret_passage_chests)
    graph.add_edge(west_hall_post_dungeon, west_hall_post_dungeon_chests)
    graph.add_edge(west_wing, west_wing_chests)
    graph.add_edge(beasts_room, beasts_room_chests)

    if not graph.reverse_rando:
        graph.add_edge(START_NODE, courtyard)
        graph.add_edge(courtyard, belles_room)
        graph.add_edge(belles_room, east_wing)
        graph.add_edge(courtyard, west_hall)
        graph.add_edge(west_hall, thresholder, RequirementEdge(battle=True))
        graph.add_edge(thresholder, dungeon)
        graph.add_edge(dungeon, secret_passage)
        graph.add_edge(secret_passage, west_hall_post_dungeon)
        graph.add_edge(west_hall_post_dungeon, west_wing)
        graph.add_edge(west_wing, beast_bonus, RequirementEdge(battle=True))
        graph.add_edge(beast_bonus, beasts_room)
        graph.add_edge(beasts_room, dark_thorn, RequirementEdge(battle=True))
        graph.add_edge(dark_thorn, rumbling_rose, RequirementEdge(battle=True))
        graph.add_edge(rumbling_rose, xaldin, RequirementEdge(battle=True))
        graph.add_edge(xaldin, data_xaldin, RequirementEdge(battle=True))
        graph.register_first_boss(dark_thorn)
        graph.register_last_story_boss(xaldin)
    else:
        graph.add_edge(START_NODE, west_hall, RequirementEdge(battle=True))
        graph.add_edge(west_hall, west_hall_post_dungeon)
        graph.add_edge(west_hall, west_wing)
        graph.add_edge(west_wing, beasts_room)
        graph.add_edge(beasts_room, rumbling_rose)
        graph.add_edge(rumbling_rose, xaldin, RequirementEdge(battle=True))
        graph.add_edge(xaldin, courtyard)
        graph.add_edge(courtyard, belles_room)
        graph.add_edge(belles_room, east_wing)
        graph.add_edge(east_wing, thresholder, RequirementEdge(battle=True))
        graph.add_edge(thresholder, dungeon)
        graph.add_edge(dungeon, secret_passage)
        graph.add_edge(secret_passage, beast_bonus, RequirementEdge(battle=True))
        graph.add_edge(beast_bonus, dark_thorn, RequirementEdge(battle=True))
        graph.add_edge(dark_thorn, data_xaldin, RequirementEdge(battle=True))
        graph.register_first_boss(xaldin)
        graph.register_last_story_boss(dark_thorn)



from dataclasses import dataclass, field
from typing import Optional

from List.configDict import itemType, locationType
from List.inventory.item import InventoryItem


@dataclass(frozen=True)
class StoryUnlock(InventoryItem):
    id: int
    name: str
    location: locationType
    visit_count: int
    type: itemType = field(init=False, default=itemType.STORYUNLOCK)


BattlefieldsOfWar = StoryUnlock(54, "Battlefields of War (Auron)", locationType.OC, visit_count=2)
SwordOfTheAncestor = StoryUnlock(55, "Sword of the Ancestor (Mulan)", locationType.LoD, visit_count=2)
BeastsClaw = StoryUnlock(59, "Beast's Claw (Beast)", locationType.BC, visit_count=2)
BoneFist = StoryUnlock(60, "Bone Fist (Jack Skellington)", locationType.HT, visit_count=2)
ProudFang = StoryUnlock(61, "Proud Fang (Simba)", locationType.PL, visit_count=2)
SkillAndCrossbones = StoryUnlock(62, "Skill and Crossbones (Jack Sparrow)", locationType.PR, visit_count=2)
Scimitar = StoryUnlock(72, "Scimitar (Aladdin)", locationType.Agrabah, visit_count=2)
WayToTheDawn = StoryUnlock(73, "Way to the Dawn (Riku)", locationType.TWTNW, visit_count=2)
IdentityDisk = StoryUnlock(74, "Identity Disk (Tron)", locationType.SP, visit_count=2)
# TournamentPoster = StoryUnlock(365, "Tournament Poster")
MembershipCard = StoryUnlock(369, "Membership Card", locationType.HB, visit_count=2)
IceCream = StoryUnlock(375, "Ice Cream", locationType.TT, visit_count=3)
# TODO(zaktherobot): leaving out picture for now, may revisit if there's no confusion that picture doesn't unlock TT2 anymore
# Picture = StoryUnlock(376, "Picture")
NaminesSketches = StoryUnlock(368, "Namine's Sketches", locationType.STT, visit_count=1)
RoyalSummons = StoryUnlock(460, "Royal Summons", locationType.DC, visit_count=2)  # Dummy 13


def all_story_unlocks() -> list[StoryUnlock]:
    return [BattlefieldsOfWar, SwordOfTheAncestor, BeastsClaw, BoneFist, ProudFang, SkillAndCrossbones, Scimitar,
            WayToTheDawn, IdentityDisk, MembershipCard, IceCream, NaminesSketches, RoyalSummons]


def all_individual_story_unlocks() -> list[StoryUnlock]:
    result: list[StoryUnlock] = []
    for unlock in all_story_unlocks():
        result.extend([unlock] * unlock.visit_count)
    return result


def story_unlock_for_location(location: locationType) -> Optional[StoryUnlock]:
    return next((unlock for unlock in all_story_unlocks() if unlock.location == location), None)



import copy
import random
from itertools import chain
from typing import Any, Optional

from Class.exceptions import HintException
from Class.itemClass import KH2Item
from Class.newLocationClass import KH2Location
from List.LvupStats import DreamWeaponOffsets
from List.ObjectiveList import KH2Objective
from List.configDict import HintType, SoraLevelOption, itemType, locationCategory, locationType
from List.inventory import ability, form, magic, misc, storyunlock, summon
from List.inventory.item import InventoryItem
from List.inventory.report import AnsemReport
from List.location import simulatedtwilighttown as stt
from Module.RandomizerSettings import RandomizerSettings
from Module.newRandomize import ItemAssignment


class CommonTrackerInfo:
    def __init__(self, settings: RandomizerSettings):
        self.hintsType = settings.hintsType
        self.generatorVersion = settings.ui_version
        tracker_includes = settings.tracker_includes
        self.important_check_list = settings.important_checks
        self.spoiler_reveal_list = settings.spoiler_reveal_checks
        self.objective_mode = settings.objective_rando
        self.emblem_mode = settings.emblems
        # remove any items that aren't enabled by settings
        if settings.promiseCharm:
            tracker_includes.append("PromiseCharm")
        if (
            itemType.OCSTONE in self.important_check_list
        ):  # questionable tracker_include. Consider more general alternative
            tracker_includes.append("extra_ics")
        if settings.antiform:
            tracker_includes.append("Anti-Form")
        if settings.dummy_forms:
            tracker_includes.append("dummy_forms")
        self.settings = tracker_includes
        self.point_values = settings.point_hint_values
        self.hintable_categories = settings.hintable_check_types
        self.dummy_forms = True if settings.dummy_forms else None
        self.progression_settings = (
            settings.progression_hint_settings if settings.progression_hints else None
        )
        self.spoiler_reveal = (
            settings.spoiler_hint_values if self.hintsType == HintType.SPOILER else None
        )
        self.reveal_mode = settings.revealMode

    def to_dict(self) -> dict[str, Any]:
        data: dict[str, Any] = {
            "hintsType": self.hintsType,
            "generatorVersion": self.generatorVersion,
            "settings": self.settings,
            "checkValue": self.point_values,
            "hintableItems": self.hintable_categories
        }
        if self.reveal_mode == "bossreports":
            data["ProgressionType"] = "Bosses"
        if self.dummy_forms is not None:
            data["dummy_forms"] = True
        if self.progression_settings is not None:
            data["ProgressionSettings"] = self.progression_settings
        if self.spoiler_reveal is not None:
            data["reveal"] = self.spoiler_reveal
        return data


class WorldItems:
    def __init__(
        self,
        location_item_tuples: list[tuple[KH2Location, KH2Item]],
        tracker_info: CommonTrackerInfo,
    ):
        hintable_worlds = HintUtils.hintable_worlds()
        item_to_vanilla_world = HintUtils.item_to_vanilla_world()

        self.world_to_item_list: dict[locationType, list[KH2Item]] = {}
        self.world_to_item_list_revealed: dict[locationType, list[KH2Item]] = {}
        for world in hintable_worlds:
            self.world_to_item_list[world] = []
            self.world_to_item_list_revealed[world] = []
        # add in the starting items
        self.world_to_item_list[locationType.Critical] = []
        self.world_to_item_list[locationType.Free] = []
        self.report_information: dict[int, dict[str, Any]] = {}
        self.proof_of_connection_world: Optional[locationType] = None
        self.proof_of_peace_world: Optional[locationType] = None
        self.proof_of_nonexistence_world: Optional[locationType] = None
        self.path_breadcrump_map: dict[locationType, set[locationType]] = {}

        self.create_level_check_data_for_tracker(location_item_tuples, tracker_info)
        for world_with_vanilla in HintUtils.world_to_vanilla_items().keys():
            if world_with_vanilla not in self.path_breadcrump_map:
                self.path_breadcrump_map[world_with_vanilla] = set()
        for i in range(1, 14):
            self.report_information[i] = {
                "FoundIn": "Garden of Assemblage"
            }  # default to found in starting
        important_checks = tracker_info.important_check_list
        spoiler_reveal_list = tracker_info.spoiler_reveal_list
        for location, item in location_item_tuples:
            if locationType.WeaponSlot in location.LocationTypes:
                continue

            inventory_item = item.item
            if isinstance(inventory_item, AnsemReport):
                world_of_location = HintUtils.location_to_tracker_world(location.LocationTypes)
                self.report_information[inventory_item.report_number]["FoundIn"] = world_of_location

            if item.ItemType in spoiler_reveal_list or item.Name in spoiler_reveal_list:
                world_of_location = HintUtils.location_to_tracker_world(location.LocationTypes)
                if world_of_location not in self.world_to_item_list_revealed:
                    raise HintException(
                        f"Something is going wrong with initializing worlds {world_of_location}"
                    )
                self.world_to_item_list_revealed[world_of_location].append(item)
            if item.ItemType in important_checks or item.Name in important_checks:
                world_of_location = HintUtils.location_to_tracker_world(location.LocationTypes)
                if world_of_location not in self.world_to_item_list:
                    raise HintException(
                        f"Something is going wrong with initializing worlds {world_of_location}"
                    )
                self.world_to_item_list[world_of_location].append(item)
                if item.ItemType is itemType.PROOF_OF_PEACE:
                    self.proof_of_peace_world = world_of_location
                elif item.ItemType is itemType.PROOF_OF_CONNECTION:
                    self.proof_of_connection_world = world_of_location
                elif item.ItemType is itemType.PROOF_OF_NONEXISTENCE:
                    self.proof_of_nonexistence_world = world_of_location
                elif item.ItemType not in [
                    itemType.KEYITEM,
                    itemType.REPORT,
                    itemType.PROMISE_CHARM,
                    itemType.OCSTONE,
                    itemType.TROPHY,
                    itemType.MANUFACTORYUNLOCK,
                    itemType.MUNNY_POUCH,
                    itemType.KEYBLADE,
                ] and item.Id not in [form.AntiForm.id]:
                    # this item could have come from any world from this list
                    for vanilla_world in item_to_vanilla_world[inventory_item]:
                        if world_of_location in hintable_worlds:
                            self.path_breadcrump_map[vanilla_world].add(world_of_location)

    def create_level_check_data_for_tracker(self, location_item_tuples, tracker_info):
        dream_weapon_matters_setting = "Dream Weapon Matters"
        dream_weapon_matters = dream_weapon_matters_setting in tracker_info.settings
        offsets = DreamWeaponOffsets()
        max_level = 1
        if SoraLevelOption.LEVEL_50 in tracker_info.settings:
            max_level = 50
        elif SoraLevelOption.LEVEL_99 in tracker_info.settings:
            max_level = 99

        self.level_checks = {}
        self.level_checks[locationCategory.VALORLEVEL] = {}
        self.level_checks[locationCategory.WISDOMLEVEL] = {}
        self.level_checks[locationCategory.LIMITLEVEL] = {}
        self.level_checks[locationCategory.MASTERLEVEL] = {}
        self.level_checks[locationCategory.FINALLEVEL] = {}
        self.level_checks[locationCategory.LEVEL] = {}
        self.level_checks[locationCategory.LEVEL]["Sword"] = {}
        self.level_checks[locationCategory.LEVEL]["Staff"] = {}
        self.level_checks[locationCategory.LEVEL]["Shield"] = {}

        important_checks = tracker_info.important_check_list
        for location, item in location_item_tuples:
            if item.ItemType in important_checks or item.Name in important_checks:
                if locationType.Level in location.LocationTypes:
                    self.level_checks[locationCategory.LEVEL]["Sword"][location.LocationId] = item.Id
                    if not dream_weapon_matters:
                        self.level_checks[locationCategory.LEVEL]["Staff"][location.LocationId] = item.Id
                        self.level_checks[locationCategory.LEVEL]["Shield"][location.LocationId] = item.Id
                    if dream_weapon_matters:
                        self.level_checks[locationCategory.LEVEL]["Staff"][offsets.get_staff_level(max_level,location.LocationId)] = item.Id
                        self.level_checks[locationCategory.LEVEL]["Shield"][offsets.get_shield_level(max_level,location.LocationId)] = item.Id

                if locationType.FormLevel in location.LocationTypes:
                    for form_category in self.level_checks.keys():
                        if form_category is location.LocationCategory:
                            self.level_checks[form_category][location.LocationId] = item.Id

    def world_to_item_ids(self) -> dict[locationType, list[int]]:
        """
        Returns a dictionary of world to a list of IDs of items in that world.
        """
        dictionary: dict[locationType, list[int]] = {}
        for world, item_list in self.world_to_item_list.items():
            dictionary[world] = [item.Id for item in item_list]
        return dictionary
    
    def item_ids_to_names(self) -> dict[int, str]:
        """
        Returns a dictionary of world to a list of IDs of items in that world.
        """
        dictionary: dict[int, str] = {}
        for _, item_list in self.world_to_item_list.items():
            for item in item_list:
                dictionary[item.Id] = item.Name
        return dictionary
    
    def revealed_item_ids_to_names(self) -> dict[int, str]:
        """
        Returns a dictionary of world to a list of IDs of items in that world.
        """
        dictionary: dict[int, str] = {}
        for _, item_list in self.world_to_item_list_revealed.items():
            for item in item_list:
                dictionary[item.Id] = item.Name
        return dictionary


class JsmarteeHintData:
    def __init__(
        self,
        world_items: WorldItems,
        world_to_hint: locationType,
    ):
        self.world = world_to_hint
        self.count = len(world_items.world_to_item_list[world_to_hint])


class PointHintData:
    def __init__(
        self,
        settings: RandomizerSettings,
        common_tracker_data: CommonTrackerInfo,
        world_items: WorldItems,
        world_to_hint: locationType,
    ):
        self.prevent_self_hinting = settings.prevent_self_hinting
        self.report_reveal_items = settings.spoiler_reveal_checks

        # self.allow_proof_hinting = settings.allow_proof_hinting
        # self.allow_report_hinting = settings.allow_report_hinting
        self.world = world_to_hint
        self.unique_items = []
        ic_list = common_tracker_data.important_check_list
        for item in world_items.world_to_item_list[world_to_hint]:
            if item.ItemType in ic_list or item.Name in ic_list:
                if item not in self.unique_items:
                    self.unique_items.append(item)

    def candidate_hints(self) -> list[KH2Item]:
        candidate_items: list[KH2Item] = []
        for item in self.unique_items:
            if item.ItemType in self.report_reveal_items:
                candidate_items.append(item)
        return candidate_items


class PathHintData:
    def __init__(
        self,
        world_items: WorldItems,
        world_to_hint: locationType,
    ):
        self.num_items = len(world_items.world_to_item_list[world_to_hint])
        self.world = world_to_hint
        world_text = HintUtils.location_hint_user_friendly_text(world_to_hint)

        poc_world = world_items.proof_of_connection_world
        points_to_connection = (poc_world and world_to_hint in world_items.path_breadcrump_map[poc_world])
        pop_world = world_items.proof_of_peace_world
        points_to_peace = (pop_world and world_to_hint in world_items.path_breadcrump_map[pop_world])
        pon_world = world_items.proof_of_nonexistence_world
        points_to_nonexistence = (pon_world and world_to_hint in world_items.path_breadcrump_map[pon_world])

        self.proof_list = (
            []
            + (["Connection"] if points_to_connection else [])
            + (["Peace"] if points_to_peace else [])
            + (["Nonexistence"] if points_to_nonexistence else [])
        )
        if len(self.proof_list) == 0:
            self.proof_list = ["none"]

        if self.num_items == 0:
            hint_text = f"{world_text} has nothing, sorry."
        elif not points_to_connection and not points_to_nonexistence and not points_to_peace:
            hint_text = f"{world_text} has no path to the light."
        elif points_to_connection and points_to_nonexistence and points_to_peace:
            hint_text = f"{world_text} has a path to all lights."
        elif points_to_connection and points_to_peace:
            hint_text = f"{world_text} is on the path to Connection and Peace."
        elif points_to_connection and points_to_nonexistence:
            hint_text = f"{world_text} is on the path to Connection and Nonexistence."
        elif points_to_nonexistence and points_to_peace:
            hint_text = f"{world_text} is on the path to Nonexistence and Peace."
        elif points_to_nonexistence:
            hint_text = f"{world_text} is on the path to Nonexistence."
        elif points_to_peace:
            hint_text = f"{world_text} is on the path to Peace."
        elif points_to_connection:
            hint_text = f"{world_text} is on the path to Connection."
        else:
            raise HintException("Invalid combination of path proof hinting")
        self.hint_text = hint_text


class HintUtils:
    @staticmethod
    def hintable_worlds() -> list[locationType]:
        return [
            locationType.Level,
            locationType.LoD,
            locationType.BC,
            locationType.HB,
            locationType.TT,
            locationType.TWTNW,
            locationType.SP,
            locationType.Atlantica,
            locationType.PR,
            locationType.OC,
            locationType.Agrabah,
            locationType.HT,
            locationType.PL,
            locationType.DC,
            locationType.HUNDREDAW,
            locationType.STT,
            locationType.FormLevel,
            locationType.Creations,
        ]

    @staticmethod
    def convert_item_assignment_to_tuple(
            item_assignment: list[ItemAssignment],
            shop_items: list[KH2Item]
    ) -> list[tuple[KH2Location, KH2Item]]:
        location_items: list[tuple[KH2Location, KH2Item]] = []
        for assignment in item_assignment:
            location = assignment.location
            if location.name() != stt.CheckLocation.StruggleLoserMedal:
                location_items.append((location, assignment.item))
                if assignment.item2 is not None:
                    location_items.append((location, assignment.item2))

        fake_shop_location = KH2Location(
            999,
            "Shop Item",
            locationCategory.CREATION,
            [locationType.SHOP, locationType.Creations],
        )
        for shop_item in shop_items:
            location_items.append((fake_shop_location, shop_item))

        return location_items

    @staticmethod
    def location_to_tracker_world(
        location_categories: list[locationType],
    ) -> locationType:
        hintable_worlds = HintUtils.hintable_worlds() + [
            locationType.Free,
            locationType.Critical,
        ]
        overlap = list(set(location_categories).intersection(hintable_worlds))
        if len(overlap) > 1:
            raise HintException(f"Location has two tracker worlds")
        elif len(overlap) == 0:
            raise HintException(f"Location has no tracker worlds")
        return overlap[0]

    @staticmethod
    def update_disabled_worlds_on_tracker(settings: RandomizerSettings) -> list[locationType]:
        exclude_list = copy.deepcopy(settings.disabledLocations)
        # If HB is off, but Transport or CoR are on, we gotta turn HB on
        if locationType.HB in exclude_list and (
            locationType.TTR not in exclude_list or locationType.CoR not in exclude_list
        ):
            exclude_list.remove(locationType.HB)

        # If OC is off, but Cups are on, gotta turn OC on
        if locationType.OC in exclude_list and (locationType.OCCups not in exclude_list):
            exclude_list.remove(locationType.OC)

        # find out if we need to turn on the creations world (default yes, this will disable it)
        if (
            locationType.SYNTH in exclude_list
            and locationType.Puzzle in exclude_list
            and not settings.shop_hintable
        ):
            exclude_list.append(locationType.Creations)

        # make sure vanilla worlds are in the tracker
        for vanilla in settings.vanillaLocations:
            if vanilla in exclude_list:
                exclude_list.remove(vanilla)

        return exclude_list

    @staticmethod
    def world_to_vanilla_items() -> dict[locationType, list[InventoryItem]]:
        """
        Returns a dictionary of worlds to a list of vanilla items in that world.
        """
        return {
            locationType.Level: [
                ability.SecondChance,
                ability.OnceMore,
            ],
            locationType.FormLevel: [
                form.ValorForm,
                form.WisdomForm,
                form.FinalForm,
                form.MasterForm,
                form.LimitForm,
            ],
            locationType.Atlantica: [magic.Blizzard],
            locationType.TWTNW: [
                magic.Magnet,
                storyunlock.WayToTheDawn,
            ],
            locationType.PR: [
                magic.Magnet,
                summon.FeatherCharm,
                storyunlock.SkillAndCrossbones,
            ],
            locationType.DC: [
                misc.TornPages,
                magic.Reflect,
                form.WisdomForm,
                storyunlock.RoyalSummons
            ],
            locationType.HUNDREDAW: [
                magic.Cure,
                misc.TornPages,
            ],
            locationType.Agrabah: [
                summon.LampCharm,
                magic.Fire,
                misc.TornPages,
                storyunlock.Scimitar,
            ],
            locationType.BC: [
                magic.Cure,
                magic.Reflect,
                storyunlock.BeastsClaw,
            ],
            locationType.TT: [
                form.ValorForm,
                form.LimitForm,
                storyunlock.IceCream,
                # storyunlock.Picture,
            ],
            locationType.SP: [
                magic.Reflect,
                storyunlock.IdentityDisk,
            ],
            locationType.HT: [
                magic.Magnet,
                storyunlock.BoneFist,
            ],
            locationType.PL: [
                misc.TornPages,
                magic.Fire,
                magic.Thunder,
                storyunlock.ProudFang,
            ],
            locationType.LoD: [
                magic.Thunder,
                misc.TornPages,
                storyunlock.SwordOfTheAncestor,
            ],
            locationType.OC: [
                magic.Thunder,
                storyunlock.BattlefieldsOfWar,
            ],
            locationType.HB: [
                magic.Fire,
                magic.Blizzard,
                summon.BaseballCharm,
                summon.UkuleleCharm,
                form.MasterForm,
                magic.Cure,
                misc.TornPages,
                storyunlock.MembershipCard,
            ],
            locationType.STT: [
                form.ValorForm,
                form.LimitForm,
                storyunlock.NaminesSketches
            ],
            locationType.Creations: [],
            locationType.Critical: [],
            locationType.Free: []
        }

    @staticmethod
    def item_to_vanilla_world() -> dict[InventoryItem, list[locationType]]:
        """
        Returns a dictionary of item to the list of vanilla worlds that contain that item.
        """
        result: dict[InventoryItem, list[locationType]] = {}
        for world, item_list in HintUtils.world_to_vanilla_items().items():
            for item in item_list:
                if item not in result:
                    result[item] = []
                result[item].append(world)
        return result

    @staticmethod
    def point_hint_report_assignment(
        settings: RandomizerSettings,
        world_items: WorldItems,
        point_data: list[PointHintData],
    ) -> dict[int, dict[str, Any]]:
        prevent_self_hinting = settings.prevent_self_hinting
        report_number_to_world_to_item = []
        for report_number in range(1, 14):
            for p in point_data:
                if (
                    not prevent_self_hinting
                    or world_items.report_information[report_number]["FoundIn"] != p.world
                ):
                    for cand in p.candidate_hints():
                        report_number_to_world_to_item.append((report_number, p.world, cand))

        def pick_item(
                input_report_num: int,
                input_world: locationType,
                input_data: list[tuple[int, locationType, KH2Item]]
        ) -> tuple[Optional[int], Optional[locationType], Optional[KH2Item]]:
            filtered_list = [
                (r, w, i) for r, w, i in input_data if w == input_world and r == input_report_num
            ]
            if len(filtered_list) == 0:
                return None, None, None
            return random.choice(filtered_list)

        def reduce_list(
                input_report_num: int,
                input_world: Optional[locationType],
                input_item: Optional[KH2Item],
                input_data: list[tuple[int, locationType, KH2Item]]
        ) -> list[tuple[int, locationType, KH2Item]]:
            filtered_list = [(r, w, i) for r, w, i in input_data if r != input_report_num]
            filtered_list = [
                (r, w, i) for r, w, i in filtered_list if input_world != w or input_item != i
            ]
            return filtered_list

        # populate the hintable world list with ones that have valid items
        hintable_worlds = [p.world for p in point_data if len(p.candidate_hints()) > 0]
        for _ in range(50):
            data = copy.deepcopy(report_number_to_world_to_item)
            report_assignments = {}
            random.shuffle(hintable_worlds)
            if len(hintable_worlds) >= 13:
                final_hintable_worlds = hintable_worlds[0:13]
            else:
                final_hintable_worlds = hintable_worlds * 13
                final_hintable_worlds = final_hintable_worlds[0:13]
            for report_num in range(1, 14):
                selected_report, selected_world, selected_item = pick_item(
                    report_num, final_hintable_worlds[report_num - 1], data
                )
                if selected_report is None:
                    break
                if (
                    selected_report != report_num
                    or selected_world != final_hintable_worlds[report_num - 1]
                ):
                    raise HintException("Improper data computation for point hints")
                report_assignments[report_num] = {
                    "World": selected_world,
                    "check": selected_item.Id,
                    "Location": world_items.report_information[report_num]["FoundIn"],
                    "itemName": selected_item.Name,
                }
                data = reduce_list(report_num, selected_world, selected_item, data)
            if len(report_assignments) == 13:
                return report_assignments

        raise HintException("Can't find valid point hint assignment")

    @staticmethod
    def jsmartee_hint_report_assignment(
        settings: RandomizerSettings,
        world_items: WorldItems,
        jsmartee_data: list[JsmarteeHintData],
    ) -> dict[int, dict[str, Any]]:
        prevent_self_hinting: bool = settings.prevent_self_hinting
        augmented_jsmartee_data = copy.deepcopy(jsmartee_data)
        # populate the hintable world list with ones that have valid items
        hintable_worlds = [j.world for j in augmented_jsmartee_data]
        # if we aren't progression hints, we need 13 hints at least
        if len(hintable_worlds) < 13:
            # if there aren't 13 worlds to hint, add the difference as zero IC worlds
            excluded_worlds = list(
                set(HintUtils.hintable_worlds()).difference(hintable_worlds)
            )
            augmented_jsmartee_data = augmented_jsmartee_data + [
                JsmarteeHintData(world_items, w)
                for w in excluded_worlds[0: (13 - len(augmented_jsmartee_data))]
            ]
            hintable_worlds = [j.world for j in augmented_jsmartee_data]
        report_number_to_world_to_count = []
        for num in range(1, 14):
            for p in augmented_jsmartee_data:
                if (
                    not prevent_self_hinting
                    or world_items.report_information[num]["FoundIn"] != p.world
                ):
                    report_number_to_world_to_count.append((num, p.world, p.count))

        def pick_world(
                input_report_num: int,
                input_world: locationType,
                input_data: list[tuple[int, locationType, int]]
        ) -> tuple[Optional[int], Optional[locationType], Optional[int]]:
            filtered_list = [
                (r, w, c) for r, w, c in input_data if w == input_world and r == input_report_num
            ]
            if len(filtered_list) == 0:
                return None, None, None
            return random.choice(filtered_list)

        def reduce_list(
                input_report_num: Optional[int],
                input_world: Optional[locationType],
                input_data: list[tuple[int, locationType, int]]
        ) -> list[tuple[int, locationType, int]]:
            filtered_list = [(r, w, c) for r, w, c in input_data if r != input_report_num]
            filtered_list = [(r, w, c) for r, w, c in filtered_list if input_world != w]
            return filtered_list

        proof_worlds = [
            world_items.proof_of_connection_world,
            world_items.proof_of_peace_world,
            world_items.proof_of_nonexistence_world,
        ]
        # validation for starting proof locations
        proof_worlds = [
            proof_world for proof_world in proof_worlds if proof_world in HintUtils.hintable_worlds()
        ]
        priority_worlds = []
        # TODO add warning about using visit locks with Jsmartee in RandomizerSettings
        if locationType.FormLevel in proof_worlds:
            # add the worlds with forms to priority list
            for world, item_list in world_items.world_to_item_list.items():
                if any([i.ItemType == itemType.FORM for i in item_list]):
                    priority_worlds.append(world)
        if locationType.HUNDREDAW in proof_worlds:
            # add the worlds with pages to priority list
            for world, item_list in world_items.world_to_item_list.items():
                if any([i.ItemType == itemType.TORN_PAGE for i in item_list]):
                    priority_worlds.append(world)
        if locationType.Atlantica in proof_worlds:
            # add the worlds with thunders and magnets to priority list
            for world, item_list in world_items.world_to_item_list.items():
                if any([i.ItemType == itemType.THUNDER for i in item_list]):
                    priority_worlds.append(world)
                if any([i.ItemType == itemType.MAGNET for i in item_list]):
                    priority_worlds.append(world)

        full_priority_list = [proof_world for proof_world in proof_worlds]
        for world in priority_worlds:
            if world not in full_priority_list:
                full_priority_list.append(world)

        for _ in range(100):
            data = copy.deepcopy(report_number_to_world_to_count)
            priorities = copy.deepcopy(full_priority_list)

            # cull down the world list to a set of 13
            selected_worlds = copy.deepcopy(priorities)
            remaining_worlds = [world for world in hintable_worlds if world not in selected_worlds]
            random.shuffle(remaining_worlds)
            selected_worlds = selected_worlds + remaining_worlds
            selected_worlds = selected_worlds[0:13]

            # sanity check, if we already failed to select all priority worlds, fail
            if any(world not in selected_worlds for world in priorities):
                continue

            report_numbers = list(range(1, 14))
            random.shuffle(report_numbers)
            report_assignments = {}
            # try to assign the proof reports
            for index, world in enumerate(selected_worlds):
                chosen_report = report_numbers[index]
                selected_report, selected_world, selected_count = pick_world(
                    chosen_report, world, data
                )
                # failed to select this report/world pair (likely self hinting)
                if selected_report is None:
                    break
                # Sanity check that we returned the same stuff
                if (
                    selected_report != chosen_report
                    or selected_world != selected_worlds[index]
                ):
                    raise HintException("Improper data computation for Jsmartee hints")
                # augment the priority list if we are hinting proof worlds
                if selected_world in proof_worlds:
                    report_world = world_items.report_information[chosen_report][
                        "FoundIn"
                    ]
                    if report_world not in priorities and report_world != "Garden of Assemblage":
                        priorities.append(report_world)
                    if any(inner_w not in selected_worlds for inner_w in priorities):
                        break
                    if len(priorities) > 13:
                        break

                report_assignments[chosen_report] = {
                    "World": selected_world,
                    "Location": world_items.report_information[chosen_report][
                        "FoundIn"
                    ],
                    "Count": selected_count,
                }
                data = reduce_list(chosen_report, selected_world, data)

            if len(report_assignments) == 13:
                return report_assignments

        raise HintException("Can't find valid Jsmartee hint assignment")

    @staticmethod
    def jsmartee_progression_hints(
        world_items: WorldItems,
        jsmartee_data: list[JsmarteeHintData],
    ) -> dict[int, dict[str, Any]]:
        # output all the data as sequential reports
        report_assignments = {}
        random.shuffle(jsmartee_data)
        for chosen_report, hint_data in enumerate(jsmartee_data):
            if chosen_report < 13:
                location = world_items.report_information[chosen_report + 1]["FoundIn"]
            else:
                location = ""
            report_assignments[chosen_report + 1] = {
                "World": hint_data.world,
                "Location": location,
                "Count": hint_data.count,
            }
        return report_assignments

    @staticmethod
    def path_hint_assignment(
        world_items: WorldItems,
        path_data: list[PathHintData],
        tracker_data: CommonTrackerInfo,
        hintable_worlds: list[locationType],
    ) -> dict[int, dict[str, Any]]:
        progression_hints = tracker_data.progression_settings is not None
        must_hint = list(
            set(
                chain(
                    world_items.path_breadcrump_map[
                        world_items.proof_of_connection_world
                    ]
                    if world_items.proof_of_connection_world
                    else [],
                    world_items.path_breadcrump_map[world_items.proof_of_peace_world]
                    if world_items.proof_of_peace_world
                    else [],
                    world_items.path_breadcrump_map[
                        world_items.proof_of_nonexistence_world
                    ]
                    if world_items.proof_of_nonexistence_world
                    else [],
                )
            )
        )
        must_hint_data = [p for p in path_data if p.world in must_hint]
        other_data = [p for p in path_data if p.world not in must_hint]
        # prioritize hinting the most items possible
        must_hint_data.sort(reverse=True, key=lambda x: x.num_items)
        other_data.sort(reverse=True, key=lambda x: x.num_items)
        all_data = must_hint_data + other_data

        for _ in range(50):
            report_assignments = {}
            report_numbers = list(range(1, 14))
            # shuffle first 13 reports
            random.shuffle(report_numbers)
            if progression_hints:
                report_numbers = report_numbers + list(range(14, len(all_data) + 1))
                # make sure all worlds with no items are last, and worlds that are enabled are before disabled ones
                all_data = [p for p in all_data if p.num_items > 0] + \
                            [p for p in all_data if p.num_items == 0 and p.world in hintable_worlds] + \
                            [p for p in all_data if p.num_items == 0 and p.world not in hintable_worlds]
            for index, report_number in enumerate(report_numbers):
                if index < 13:
                    location = world_items.report_information[report_number]["FoundIn"]
                else:
                    location = ""
                report_assignments[report_number] = {
                    "HintedWorld": all_data[index].world,
                    "Location": location,
                    "ProofPath": all_data[index].proof_list,
                    "Text": all_data[index].hint_text,
                }
            if len(report_numbers) == len(report_assignments):
                return report_assignments

        raise HintException("Can't find valid Path hint assignment")

    @staticmethod
    def spoiler_hint_assignment(
        settings: RandomizerSettings,
        tracker_data: CommonTrackerInfo,
        world_items: WorldItems,
        hintable_worlds: list[locationType],
    ) -> dict[int, dict[str, Any]]:
        worlds_to_hint = copy.deepcopy(hintable_worlds)
        if tracker_data.progression_settings is not None:
            # report locations don't matter, but we'll populate the report locations anyway
            data = {}
            random.shuffle(worlds_to_hint)
            for index, w in enumerate(worlds_to_hint):
                if index < 13:
                    location = world_items.report_information[index + 1]["FoundIn"]
                else:
                    location = ""
                data[index + 1] = {"World": w, "Location": location}
            return data

        for _ in range(50):
            random.shuffle(worlds_to_hint)
            selected_worlds = worlds_to_hint[0:13]
            data = {}
            for index, w in enumerate(selected_worlds):
                if (
                    not settings.prevent_self_hinting
                    or world_items.report_information[index + 1]["FoundIn"] != w
                ):
                    data[index + 1] = {
                        "World": w,
                        "Location": world_items.report_information[index + 1]["FoundIn"],
                    }
            if len(data) == 13:
                return data

        raise HintException("Can't find valid spoiler hint assignment")

    @staticmethod
    def location_hint_user_friendly_text(location: locationType) -> str:
        """
        Returns a user-friendly variation of the location's name for use in hints.
        """
        if location == locationType.Level:
            return "Sora's Heart"
        elif location == locationType.TWTNW:
            return "TWTNW"
        elif location == locationType.DC:
            return "Disney Castle"
        elif location == locationType.HUNDREDAW:
            return "100 Acre"
        else:
            return location.value



from dataclasses import dataclass

from List.configDict import itemType
from List.inventory.item import InventoryItem


@dataclass(frozen=True)
class Proof(InventoryItem):
    id: int
    name: str
    type: itemType


ProofOfConnection = Proof(593, "Proof of Connection", itemType.PROOF_OF_CONNECTION)
ProofOfNonexistence = Proof(594, "Proof of Nonexistence", itemType.PROOF_OF_NONEXISTENCE)
ProofOfPeace = Proof(595, "Proof of Peace", itemType.PROOF_OF_PEACE)


def proof_item_types() -> list[itemType]:
    return [itemType.PROOF_OF_CONNECTION, itemType.PROOF_OF_NONEXISTENCE, itemType.PROOF_OF_PEACE]



from copy import deepcopy
from typing import Any, Optional, Iterator
from zipfile import ZipFile
from Module.resources import resource_path

import yaml, re

Asset = dict[str, Any]


def write_yaml_to_zip_file(zip_file: ZipFile, name: str, data, sort_keys: bool):
    zip_file.writestr(name, yaml.dump(data, line_break="\r\n", sort_keys=sort_keys))


def write_unicode_yaml_to_zip_file(zip_file: ZipFile, name: str, data, sort_keys: bool):
    yaml_string = yaml.dump(data, line_break="\r\n", sort_keys=sort_keys)
    yaml_string = re.sub(
        r"en: ([a-zA-Z0-9\\]+)", r'en: "\1"', yaml_string
    )  # surround text of the journal with double quotes to allow for automatic unicode conversion
    yaml_string = yaml_string.replace("NEWLINE", "\\n")
    zip_file.writestr(name, yaml_string)


class ModYml:
    """Builder for a mod.yml file to be used within an OpenKH Mods Manager mod."""

    def __init__(self, title: str, description: Optional[str]):
        self.data: dict[str, Any] = {"title": title}
        if description is not None:
            self.data["description"] = description
        self.data["assets"] = []

    def add_assets(self, assets: list[Asset]):
        existing = self.data["assets"]
        # Using deepcopy to avoid YAML anchors and aliases being used
        for asset in assets:
            existing.append(deepcopy(asset))

    def add_asset(self, asset: Asset):
        self.add_assets([asset])

    def find_assets(self, asset_name: str) -> Iterator[Asset]:
        return (asset for asset in self.data["assets"] if asset["name"] == asset_name)

    def find_asset(self, asset_name: str) -> Optional[Asset]:
        return next(self.find_assets(asset_name), None)

    def add_asset_source(self, asset_name: str, source: dict[str, Any]):
        asset = self.find_asset(asset_name)
        if asset is None:
            raise Exception(f"Unable to find {asset_name} in the mod being built")
        else:
            # Using deepcopy to avoid YAML anchors and aliases being used
            asset["source"].append(deepcopy(source))

    def write_to_zip_file(self, zip_file: ZipFile):
        write_yaml_to_zip_file(zip_file, "mod.yml", self.data, sort_keys=False)


class Bonuses:
    """YAML builder for listpatch for bonuses. See https://openkh.dev/kh2/file/type/00battle.html#bons."""

    def __init__(self, source_name: str):
        self.data: dict[int, dict[str, dict[str, int]]] = {}
        self.source_name = source_name

    def add_bonus(
        self,
        reward_id: int,
        character_name: str,
        character_id: int,
        hp_increase: int,
        mp_increase: int,
        drive_gauge_increase: int,
        item_slot_upgrade: int,
        accessory_slot_upgrade: int,
        armor_slot_upgrade: int,
        bonus_item_1: int,
        bonus_item_2: int,
        padding: int,
    ):
        if reward_id not in self.data:
            self.data[reward_id] = {}
        self.data[reward_id][character_name] = {
            "RewardId": reward_id,
            "CharacterId": character_id,
            "HpIncrease": hp_increase,
            "MpIncrease": mp_increase,
            "DriveGaugeUpgrade": drive_gauge_increase,
            "ItemSlotUpgrade": item_slot_upgrade,
            "AccessorySlotUpgrade": accessory_slot_upgrade,
            "ArmorSlotUpgrade": armor_slot_upgrade,
            "BonusItem1": bonus_item_1,
            "BonusItem2": bonus_item_2,
            "Padding": padding,
            # "Description": ?,
            # "Unknown0c": ?,
        }

    def write_to_zip_file(self, zip_file: ZipFile):
        write_yaml_to_zip_file(zip_file, self.source_name, self.data, sort_keys=True)


class FormLevels:
    """YAML builder for listpatch for form levels. See https://openkh.dev/kh2/file/type/00battle.html#fmlv."""

    def __init__(self, source_name: str):
        self.data: dict[str, list[dict[str, int]]] = {}
        self.source_name = source_name

    def add_form_level(
        self,
        form_name: str,
        form_id: int,
        form_level: int,
        ability: int,
        experience: int,
        growth_ability_level: int,
    ):
        if form_name not in self.data:
            self.data[form_name] = []
        self.data[form_name].append(
            {
                "FormId": form_id,
                "FormLevel": form_level,
                "Experience": experience,
                "Ability": ability,
                "GrowthAbilityLevel": growth_ability_level,
            }
        )

    def turn_off_anti(self):
        self.data["Antiform"] = []
        self.data["Antiform"].append(
            {
                "FormId": 6,
                "FormLevel": 1,
                "Experience": 0,
                "Ability": 0,
                "GrowthAbilityLevel": 0,
            }
        )
        self.data["Antiform"].append(
            {
                "FormId": 6,
                "FormLevel": 2,
                "Experience": 0,
                "Ability": 0,
                "GrowthAbilityLevel": 0,
            }
        )
        self.data["Antiform"].append(
            {
                "FormId": 6,
                "FormLevel": 3,
                "Experience": 0,
                "Ability": 0,
                "GrowthAbilityLevel": 0,
            }
        )

    def write_to_zip_file(self, zip_file: ZipFile):
        write_yaml_to_zip_file(zip_file, self.source_name, self.data, sort_keys=False)


class Items:
    """YAML builder for listpatch for items. See https://openkh.dev/kh2/file/type/03system.html#item."""

    def __init__(self, source_name: str):
        self.data: dict[str, list[dict[str, int]]] = {
            "Stats": [],
            "Items": [],
        }
        self.source_name = source_name

    def add_stats(
        self,
        location_id: int,
        attack: int,
        magic: int,
        defense: int,
        ability: int,
        ability_points: int,
        unknown_08: int,
        fire_resistance: int,
        ice_resistance: int,
        lightning_resistance: int,
        dark_resistance: int,
        unknown_0d: int,
        general_resistance: int,
        unknown: int,
    ):
        self.data["Stats"].append(
            {
                "Id": location_id,
                "Attack": attack,
                "Magic": magic,
                "Defense": defense,
                "Ability": ability,
                "AbilityPoints": ability_points,
                "Unknown08": unknown_08,
                "FireResistance": fire_resistance,
                "IceResistance": ice_resistance,
                "LightningResistance": lightning_resistance,
                "DarkResistance": dark_resistance,
                "Unknown0d": unknown_0d,
                "GeneralResistance": general_resistance,
                "Unknown": unknown,
            }
        )

    def add_item(
        self,
        item_id: int,
        item_type: str,
        flag_0: int,
        flag_1: int,
        rank: str,
        stat_entry: int,
        name: int,
        description: int,
        shop_buy: int,
        shop_sell: int,
        command: int,
        slot: int,
        picture: int,
        icon_1: int,
        icon_2: int,
    ):
        self.data["Items"].append(
            {
                "Id": item_id,
                "Type": item_type,
                "Flag0": flag_0,
                "Flag1": flag_1,
                "Rank": rank,
                "StatEntry": stat_entry,
                "Name": name,
                "Description": description,
                "ShopBuy": shop_buy,
                "ShopSell": shop_sell,
                "Command": command,
                "Slot": slot,
                "Picture": picture,
                "Icon1": icon_1,
                "Icon2": icon_2,
            }
        )

    def write_to_zip_file(self, zip_file: ZipFile):
        write_yaml_to_zip_file(zip_file, self.source_name, self.data, sort_keys=False)


class LevelUps:
    """YAML builder for listpatch for level ups. See https://openkh.dev/kh2/file/type/00battle.html#lvup."""

    def __init__(self, source_name: str):
        self.data: dict[str, dict[int, dict[str, Any]]] = {"Sora": {}, "Donald": {}, "Goofy": {}, "PingMulan": {}, "Beast": {}, "Sparrow": {}, "Aladdin": {}, "Jack": {}, "Auron": {}, "Simba": {}, "Tron": {}, "Riku": {},}
        self.source_name = source_name
        with open(resource_path("static/LvupList.yml"), "r") as file:
            list_data = yaml.safe_load(file)
        self.yaml_list_data = list_data

    def add_sora_level(
        self,
        level: int,
        experience: int,
        strength: int,
        magic: int,
        defense: int,
        ap: int,
        sword_ability: int,
        shield_ability: int,
        staff_ability: int,
        padding: int,
    ):
        self.data["Sora"][level] = {
            "Character": "Sora",
            "Level": level,
            "Exp": experience,
            "Strength": strength,
            "Magic": magic,
            "Defense": defense,
            "Ap": ap,
            "SwordAbility": sword_ability,
            "ShieldAbility": shield_ability,
            "StaffAbility": staff_ability,
            "Padding": padding,
        }

    def add_companion_level(
        self,
        level: int,
        experience: int,
        strength: int,
        magic: int,
        defense: int,
        ap: int,
        companion_name: str,
        padding: int,
    ):
        self.data[companion_name][level] = {
            "Character": companion_name,
            "Level": level,
            "Exp": experience,
            "Strength": strength,
            "Magic": magic,
            "Defense": defense,
            "Ap": ap,
            "SwordAbility": 0,
            "ShieldAbility": 0,
            "StaffAbility": 0,
            "Padding": padding,
        }

    def get_companion_levels(self, companion_name: str):
        for entry in self.yaml_list_data:
            if entry == companion_name:
                return self.yaml_list_data[entry]

    def write_to_zip_file(self, zip_file: ZipFile):
        write_yaml_to_zip_file(zip_file, self.source_name, self.data, sort_keys=True)


class Messages:
    """YAML builder for listpatch for items. See the kh2msg section at https://openkh.dev/tool/GUI.ModsManager."""

    def __init__(self, source_name: str, unicode_output: bool = False):
        self.data: list[dict[str, Any]] = []
        self.source_name = source_name
        self.unicode_output = unicode_output

    def add_message(
        self, message_id: int, en: Optional[str] = None, jp: Optional[str] = None
    ):
        entry = {"id": message_id}
        if en is not None:
            entry["en"] = en
        if jp is not None:
            entry["jp"] = jp
        self.data.append(entry)

    def write_to_zip_file(self, zip_file: ZipFile):
        if not self.unicode_output:
            write_yaml_to_zip_file(
                zip_file, self.source_name, self.data, sort_keys=False
            )
        else:
            write_unicode_yaml_to_zip_file(
                zip_file, self.source_name, self.data, sort_keys=False
            )


class PlayerParams:
    """YAML builder for listpatch for player params. See https://openkh.dev/kh2/file/type/00battle.html#plrp."""

    def __init__(self, source_name: str):
        self.data: list[dict] = []
        self.source_name = source_name

    def add_player(
        self,
        character_id: int,
        identifier: int,
        hp: int,
        mp: int,
        ap: int,
        armor_slot_max: int,
        accessory_slot_max: int,
        item_slot_max: int,
        items: list[int],
        padding: list[int],
    ):
        self.data.append(
            {
                "Character": character_id,
                "Id": identifier,
                "Hp": hp,
                "Mp": mp,
                "Ap": ap,
                "ArmorSlotMax": armor_slot_max,
                "AccessorySlotMax": accessory_slot_max,
                "ItemSlotMax": item_slot_max,
                "Items": items,
                "Padding": padding,
            }
        )

    def write_to_zip_file(self, zip_file: ZipFile):
        write_yaml_to_zip_file(zip_file, self.source_name, self.data, sort_keys=False)


class PrizeTable:
    """YAML builder for listpatch for prize table. See https://openkh.dev/kh2/file/type/00battle.html#przt."""

    def __init__(self, source_name: str):
        self.data: list[dict[str, int]] = []
        self.source_name = source_name

    def has_entries(self):
        return len(self.data) > 0

    def add_prize(
        self,
        identifier: int,
        small_hp_orbs: int,
        big_hp_orbs: int,
        big_money_orbs: int,
        medium_money_orbs: int,
        small_money_orbs: int,
        small_mp_orbs: int,
        big_mp_orbs: int,
        small_drive_orbs: int,
        big_drive_orbs: int,
        item_1: int,
        item_1_percentage: int,
        item_2: int,
        item_2_percentage: int,
        item_3: int,
        item_3_percentage: int,
    ):
        self.data.append(
            {
                "Id": identifier,
                "SmallHpOrbs": small_hp_orbs,
                "BigHpOrbs": big_hp_orbs,
                "BigMoneyOrbs": big_money_orbs,
                "MediumMoneyOrbs": medium_money_orbs,
                "SmallMoneyOrbs": small_money_orbs,
                "SmallMpOrbs": small_mp_orbs,
                "BigMpOrbs": big_mp_orbs,
                "SmallDriveOrbs": small_drive_orbs,
                "BigDriveOrbs": big_drive_orbs,
                "Item1": item_1,
                "Item1Percentage": item_1_percentage,
                "Item2": item_2,
                "Item2Percentage": item_2_percentage,
                "Item3": item_3,
                "Item3Percentage": item_3_percentage,
            }
        )

    def write_to_zip_file(self, zip_file: ZipFile):
        write_yaml_to_zip_file(zip_file, self.source_name, self.data, sort_keys=False)


class Treasures:
    """YAML builder for listpatch for treasures. See https://openkh.dev/kh2/file/type/03system.html#trsr."""

    def __init__(self, source_name: str):
        self.data: dict[int, dict[str, int]] = {}
        self.source_name = source_name

    def add_treasure(self, location_id: int, item_id: int):
        self.data[location_id] = {"ItemId": item_id}

    def write_to_zip_file(self, zip_file: ZipFile):
        write_yaml_to_zip_file(zip_file, self.source_name, self.data, sort_keys=True)


class ATKPObject:
    def __init__(
        self,
        SubId: int,
        Id: int,
        Type: str,
        CriticalAdjust: int,
        Power: int,
        Team: int,
        Element: int,
        EnemyReaction: int,
        EffectOnHit: int,
        KnockbackStrength1: int,
        KnockbackStrength2: int,
        Unknown: int,
        Flags,
        RefactSelf: str,
        RefactOther: str,
        ReflectedMotion: int,
        ReflectedHitBack: int,
        ReflectAction: int,
        ReflectHitSound: int,
        ReflectRC: int,
        ReflectRange: int,
        ReflectAngle: int,
        DamageEffect: int,
        Switch: int,
        Interval: int,
        FloorCheck: int,
        DriveDrain: int,
        RevengeDamage: int,
        AttackTrReaction: str,
        ComboGroup: int,
        RandomEffect: int,
        Kind,
        HPDrain: int,
    ):
        self.SubId = SubId
        self.Id = Id
        self.Type = Type
        self.CriticalAdjust = CriticalAdjust
        self.Power = Power
        self.Team = Team
        self.Element = Element
        self.EnemyReaction = EnemyReaction
        self.EffectOnHit = EffectOnHit
        self.KnockbackStrength1 = KnockbackStrength1
        self.KnockbackStrength2 = KnockbackStrength2
        self.Unknown = Unknown
        self.Flags = Flags
        self.RefactSelf = RefactSelf
        self.RefactOther = RefactOther
        self.ReflectedMotion = ReflectedMotion
        self.ReflectedHitBack = ReflectedHitBack
        self.ReflectAction = ReflectAction
        self.ReflectHitSound = ReflectHitSound
        self.ReflectRC = ReflectRC
        self.ReflectRange = ReflectRange
        self.ReflectAngle = ReflectAngle
        self.DamageEffect = DamageEffect
        self.Switch = Switch
        self.Interval = Interval
        self.FloorCheck = FloorCheck
        self.DriveDrain = DriveDrain
        self.RevengeDamage = RevengeDamage
        self.AttackTrReaction = AttackTrReaction
        self.ComboGroup = ComboGroup
        self.RandomEffect = RandomEffect
        self.Kind = Kind
        self.HPDrain = HPDrain


class AttackEntriesOrganizer:
    def __init__(self, source_name: str):
        self.data: list[dict] = []
        self.source_name = source_name
        with open(resource_path("static/AtkpList.yml"), "r") as file:
            list_data = yaml.safe_load(file)
        self.yaml_list_data = list_data

    def convert_atkp_object_to_dict_and_add_to_data(self, atkp_object: ATKPObject):
        self.data.append(
            {
                "SubId": atkp_object.SubId,
                "Id": atkp_object.Id,
                "Type": atkp_object.Type,
                "CriticalAdjust": atkp_object.CriticalAdjust,
                "Power": atkp_object.Power,
                "Team": atkp_object.Team,
                "Element": atkp_object.Element,
                "EnemyReaction": atkp_object.EnemyReaction,
                "EffectOnHit": atkp_object.EffectOnHit,
                "KnockbackStrength1": atkp_object.KnockbackStrength1,
                "KnockbackStrength2": atkp_object.KnockbackStrength2,
                "Unknown": atkp_object.Unknown,
                "Flags": atkp_object.Flags,
                "RefactSelf": atkp_object.RefactSelf,
                "RefactOther": atkp_object.RefactOther,
                "ReflectedMotion": atkp_object.ReflectedMotion,
                "ReflectedHitBack": atkp_object.ReflectedHitBack,
                "ReflectAction": atkp_object.ReflectAction,
                "ReflectHitSound": atkp_object.ReflectHitSound,
                "ReflectRC": atkp_object.ReflectRC,
                "ReflectRange": atkp_object.ReflectRange,
                "ReflectAngle": atkp_object.ReflectAngle,
                "DamageEffect": atkp_object.DamageEffect,
                "Switch": atkp_object.Switch,
                "Interval": atkp_object.Interval,
                "FloorCheck": atkp_object.FloorCheck,
                "DriveDrain": atkp_object.DriveDrain,
                "RevengeDamage": atkp_object.RevengeDamage,
                "AttackTrReaction": atkp_object.AttackTrReaction,
                "ComboGroup": atkp_object.ComboGroup,
                "RandomEffect": atkp_object.RandomEffect,
                "Kind": atkp_object.Kind,
                "HPDrain": atkp_object.HPDrain,
            }
        )

    def write_to_zip_file(self, zip_file: ZipFile):
        write_yaml_to_zip_file(zip_file, self.source_name, self.data, sort_keys=False)

    def has_entries(self):
        return len(self.data) > 0

    def attack_entry_constructor(self, values):
        return ATKPObject(
            values["SubId"],
            values["Id"],
            values["Type"],
            values["CriticalAdjust"],
            values["Power"],
            values["Team"],
            values["Element"],
            values["EnemyReaction"],
            values["EffectOnHit"],
            values["KnockbackStrength1"],
            values["KnockbackStrength2"],
            values["Unknown"],
            values["Flags"],
            values["RefactSelf"],
            values["RefactOther"],
            values["ReflectedMotion"],
            values["ReflectedHitBack"],
            values["ReflectAction"],
            values["ReflectHitSound"],
            values["ReflectRC"],
            values["ReflectRange"],
            values["ReflectAngle"],
            values["DamageEffect"],
            values["Switch"],
            values["Interval"],
            values["FloorCheck"],
            values["DriveDrain"],
            values["RevengeDamage"],
            values["AttackTrReaction"],
            values["ComboGroup"],
            values["RandomEffect"],
            values["Kind"],
            values["HPDrain"],
        )

    def get_attack_using_ids(self, SubId, Id):
        for attack_entry in self.yaml_list_data:
            if attack_entry["SubId"] == SubId and attack_entry["Id"] == Id:
                return self.attack_entry_constructor(attack_entry)

    # Used specifically for entries that have the same Id and SubId
    def get_attack_using_ids_plus_switch(self, SubId, Id, Switch):
        for attack_entry in self.yaml_list_data:
            if (
                attack_entry["SubId"] == SubId
                and attack_entry["Id"] == Id
                and attack_entry["Switch"] == Switch
            ):
                return self.attack_entry_constructor(attack_entry)



import os
import sys


def resource_path(relative_path: str) -> str:
    """ Get absolute path to resource, works for dev and for PyInstaller """
    base_path = getattr(sys, '_MEIPASS', os.path.dirname(os.path.abspath(__file__))[:-6])
    return os.path.join(base_path, relative_path)



from dataclasses import dataclass, field

from List.configDict import itemType
from List.inventory.item import InventoryItem


@dataclass(frozen=True)
class SynthItem(InventoryItem):
    id: int
    name: str
    type: itemType = field(init=False, default=itemType.SYNTH)


DarkShard = SynthItem(280, "Dark Shard")
DarkStone = SynthItem(281, "Dark Stone")
DarkGem = SynthItem(282, "Dark Gem")
DarkCrystal = SynthItem(283, "Dark Crystal")
BlazingShard = SynthItem(317, "Blazing Shard")
BlazingStone = SynthItem(318, "Blazing Stone")
BlazingGem = SynthItem(319, "Blazing Gem")
BlazingCrystal = SynthItem(320, "Blazing Crystal")
LightningShard = SynthItem(325, "Lightning Shard")
LightningStone = SynthItem(326, "Lightning Stone")
LightningGem = SynthItem(327, "Lightning Gem")
LightningCrystal = SynthItem(328, "Lightning Crystal")
PowerShard = SynthItem(329, "Power Shard")
PowerStone = SynthItem(330, "Power Stone")
PowerGem = SynthItem(331, "Power Gem")
PowerCrystal = SynthItem(332, "Power Crystal")
LucidShard = SynthItem(333, "Lucid Shard")
LucidStone = SynthItem(334, "Lucid Stone")
LucidGem = SynthItem(335, "Lucid Gem")
LucidCrystal = SynthItem(336, "Lucid Crystal")
DenseShard = SynthItem(337, "Dense Shard")
DenseStone = SynthItem(338, "Dense Stone")
DenseGem = SynthItem(339, "Dense Gem")
DenseCrystal = SynthItem(340, "Dense Crystal")
TwilightShard = SynthItem(341, "Twilight Shard")
TwilightStone = SynthItem(342, "Twilight Stone")
TwilightGem = SynthItem(343, "Twilight Gem")
TwilightCrystal = SynthItem(344, "Twilight Crystal")
MythrilShard = SynthItem(345, "Mythril Shard")
MythrilStone = SynthItem(346, "Mythril Stone")
MythrilGem = SynthItem(347, "Mythril Gem")
MythrilCrystal = SynthItem(348, "Mythril Crystal")
BrightShard = SynthItem(349, "Bright Shard")
BrightStone = SynthItem(350, "Bright Stone")
BrightGem = SynthItem(351, "Bright Gem")
BrightCrystal = SynthItem(352, "Bright Crystal")
EnergyShard = SynthItem(353, "Energy Shard")
EnergyStone = SynthItem(354, "Energy Stone")
EnergyGem = SynthItem(355, "Energy Gem")
EnergyCrystal = SynthItem(356, "Energy Crystal")
SerenityShard = SynthItem(357, "Serenity Shard")
SerenityStone = SynthItem(358, "Serenity Stone")
SerenityGem = SynthItem(359, "Serenity Gem")
SerenityCrystal = SynthItem(360, "Serenity Crystal")
OrichalcumPlus = SynthItem(361, "Orichalcum+")
Orichalcum = SynthItem(377, "Orichalcum")
FrostShard = SynthItem(378, "Frost Shard")
FrostStone = SynthItem(379, "Frost Stone")
FrostGem = SynthItem(380, "Frost Gem")
FrostCrystal = SynthItem(381, "Frost Crystal")
RemembranceShard = SynthItem(576, "Remembrance Shard")
RemembranceStone = SynthItem(577, "Remembrance Stone")
RemembranceGem = SynthItem(578, "Remembrance Gem")
RemembranceCrystal = SynthItem(579, "Remembrance Crystal")
TranquilityShard = SynthItem(580, "Tranquility Shard")
TranquilityStone = SynthItem(581, "Tranquility Stone")
TranquilityGem = SynthItem(582, "Tranquility Gem")
TranquilityCrystal = SynthItem(583, "Tranquility Crystal")
LostIllusion = SynthItem(584, "Lost Illusion")
ManifestIllusion = SynthItem(585, "Manifest Illusion")



# Adapted from https://cs.stackexchange.com/a/65744

import string

_default_base = string.digits + string.ascii_letters


def v2r(n: int, base: str = _default_base) -> str:  # value to representation
    """Convert a positive number to its digit representation in a custom base."""
    if n == 0:
        return base[0]
    b = len(base)
    digits = ''
    while n > 0:
        digits = base[n % b] + digits
        n = n // b
    return digits


def r2v(digits: str, base: str = _default_base) -> int:  # representation to value
    """Compute the number represented by string 'digits' in a custom base."""
    b = len(base)
    n = 0
    for d in digits:
        n = b * n + base[:b].index(d)
    return n



from dataclasses import dataclass, field

from List.configDict import itemType
from List.inventory.item import InventoryItem


@dataclass(frozen=True)
class Recipe(InventoryItem):
    id: int
    name: str
    type: itemType = field(init=False, default=itemType.RECIPE)


MegaRecipe = Recipe(382, "Mega-Recipe")
StarRecipe = Recipe(449, "Star Recipe")
RecoveryRecipe = Recipe(450, "Recovery Recipe")
SkillRecipe = Recipe(451, "Skill Recipe")
GuardRecipe = Recipe(452, "Guard Recipe")
RoadToDiscovery = Recipe(464, "Road to Discovery")
StrengthBeyondStrength = Recipe(465, "Strength Beyond Strength")
BookOfShadows = Recipe(466, "Book of Shadows")
CloakedThunder = Recipe(467, "Cloaked Thunder")
EternalBlossom = Recipe(468, "Eternal Blossom")
RareDocument = Recipe(469, "Rare Document")
StyleRecipe = Recipe(475, "Style Recipe")
MoonRecipe = Recipe(476, "Moon Recipe")
QueenRecipe = Recipe(477, "Queen Recipe")
KingRecipe = Recipe(478, "King Recipe")
UltimateRecipe = Recipe(479, "Ultimate Recipe")



from enum import Enum
from typing import Optional, Union, Callable

from altgraph.Graph import Graph

from Class.exceptions import GeneratorException
from Class.newLocationClass import KH2Location
from List.configDict import locationType, itemType, locationCategory
from List.inventory.item import InventoryItem
from Module.RandomizerSettings import RandomizerSettings

RequirementFunction = Callable[[list[int]], bool]


START_NODE = "Starting"


class DefaultLogicGraph:
    def __init__(self, node_ids : Enum = None):
        # makes a dictionary of all unique ordered pairs
        self.logic = {}
        self.logic[START_NODE] = {}
        if node_ids is not None:
            for i in list(node_ids):
                self.logic[START_NODE][i] = lambda inv : True
                self.logic[i] = {}
                for j in list(node_ids):
                    if i!=j:
                        self.logic[i][j] = lambda inv : True
        
    def has_rule(self,node1,node2):
        if node1 in self.logic and node2 in self.logic[node1]:
            return True
        return False

    def merge(self, other: 'DefaultLogicGraph'):
        for start_node, out_edges in other.logic.items():
            if start_node not in self.logic:
                self.logic[start_node] = {}
            for end_node, rule in out_edges.items():
                self.logic[start_node][end_node] = rule

class LocationNode:

    def __init__(self, in_loc: Optional[list[KH2Location]] = None):
        if in_loc is None:
            self.locations = []
        else:
            self.locations = in_loc


class RequirementEdge:

    def __init__(self, req: Optional[RequirementFunction] = None, strict=True, battle=False):
        self.requirement = req
        self.strict = strict
        self.battle = battle


class LocationGraphBuilder:

    def __init__(self, graph: Graph, reverse_rando: bool, settings: RandomizerSettings):
        self.graph = graph
        self.keyblades_unlock_chests = settings.keyblades_unlock_chests
        self.reverse_rando = reverse_rando
        self.settings = settings
        self.logic_graph = DefaultLogicGraph()
        self.pending_locations_by_name: dict[str, KH2Location] = {}
        self.pending_first_boss_nodes: list[str] = []
        self.pending_last_story_boss_nodes: list[str] = []
        self.pending_superboss_nodes: list[str] = []
    
    def add_logic(self,added_logic_graph : DefaultLogicGraph):
        self.logic_graph.merge(added_logic_graph)

    def add_location(self, name: str, locations: list[KH2Location]) -> str:
        self.graph.add_node(name, LocationNode(locations))
        for location in locations:
            location_name = location.name()
            if location_name in self.pending_locations_by_name:
                raise GeneratorException(f"Location {location_name} is already in the graph")
            else:
                self.pending_locations_by_name[location_name] = location
        return name

    def add_edge(self, node1: str, node2: str, requirement: Optional[RequirementEdge] = None):
        if requirement is None:
            requirement = RequirementEdge()
        if self.logic_graph.has_rule(node1,node2) and requirement.requirement is None:
            requirement.requirement = self.logic_graph.logic[node1][node2]
        self.graph.add_edge(node1, node2, requirement)

    def register_first_boss(self, node: str):
        self.pending_first_boss_nodes.append(node)

    def register_last_story_boss(self, node: str):
        self.pending_last_story_boss_nodes.append(node)

    def register_superboss(self, node: str):
        self.pending_superboss_nodes.append(node)


def _make_location(
        loc_id: int,
        description: str,
        category: locationCategory,
        loc_types: Union[locationType, list[locationType]],
        invalid_checks: Optional[list[itemType]] = None,
        vanilla: Optional[Union[InventoryItem, list[InventoryItem]]] = None
) -> KH2Location:
    if isinstance(loc_types, locationType):
        loc_types = [loc_types]
    if invalid_checks is None:
        invalid_checks = []
    vanilla_items: list[int] = []
    if vanilla is not None:
        if isinstance(vanilla, InventoryItem):
            vanilla_items.append(vanilla.id)
        else:
            for item in vanilla:
                vanilla_items.append(item.id)
    return KH2Location(loc_id, description, category, loc_types, invalid_checks, vanilla_items)


def chest(
        loc_id: int,
        description: str,
        loc_types: Union[locationType, list[locationType]],
        invalid_checks: Optional[list[itemType]] = None,
        vanilla: Optional[Union[InventoryItem, list[InventoryItem]]] = None
) -> KH2Location:
    return _make_location(loc_id, description, locationCategory.CHEST, loc_types, invalid_checks, vanilla)


def popup(
        loc_id: int,
        description: str,
        loc_types: Union[locationType, list[locationType]],
        invalid_checks: Optional[list[itemType]] = None,
        vanilla: Optional[Union[InventoryItem, list[InventoryItem]]] = None
) -> KH2Location:
    return _make_location(loc_id, description, locationCategory.POPUP, loc_types, invalid_checks, vanilla)


def item_bonus(
        loc_id: int,
        description: str,
        loc_types: Union[locationType, list[locationType]],
        invalid_checks: Optional[list[itemType]] = None,
        vanilla: Optional[Union[InventoryItem, list[InventoryItem]]] = None
) -> KH2Location:
    return _make_location(loc_id, description, locationCategory.ITEMBONUS, loc_types, invalid_checks, vanilla)


def stat_bonus(
        loc_id: int,
        description: str,
        loc_types: Union[locationType, list[locationType]],
        invalid_checks: Optional[list[itemType]] = None,
        vanilla: Optional[Union[InventoryItem, list[InventoryItem]]] = None
) -> KH2Location:
    return _make_location(loc_id, description, locationCategory.STATBONUS, loc_types, invalid_checks, vanilla)


def hybrid_bonus(
        loc_id: int,
        description: str,
        loc_types: Union[locationType, list[locationType]],
        invalid_checks: Optional[list[itemType]] = None,
        vanilla: Optional[Union[InventoryItem, list[InventoryItem]]] = None
) -> KH2Location:
    return _make_location(loc_id, description, locationCategory.HYBRIDBONUS, loc_types, invalid_checks, vanilla)


def double_bonus(
        loc_id: int,
        description: str,
        loc_types: Union[locationType, list[locationType]],
        invalid_checks: Optional[list[itemType]] = None,
        vanilla: Optional[Union[InventoryItem, list[InventoryItem]]] = None
) -> KH2Location:
    return _make_location(loc_id, description, locationCategory.DOUBLEBONUS, loc_types, invalid_checks, vanilla)



from collections import namedtuple

ShieldStaff = namedtuple("ShieldStaff",["shield","staff"])

# Indexing other dream weapon level up reward locations based on sword
class DreamWeaponOffsets:
    def __init__(self):
        self.max_level_50 = {}
        self.max_level_99 = {}
        for i in range(1,100):
            self.max_level_50[i] = None
            self.max_level_99[i] = None
    
        self.max_level_50[2] = ShieldStaff(4,10)
        self.max_level_50[4] = ShieldStaff(10,2)
        self.max_level_50[7] = ShieldStaff(7,7)
        self.max_level_50[9] = ShieldStaff(14,23)
        self.max_level_50[10] = ShieldStaff(2,4)
        self.max_level_50[12] = ShieldStaff(12,12)
        self.max_level_50[14] = ShieldStaff(23,9)
        self.max_level_50[15] = ShieldStaff(17,34)
        self.max_level_50[17] = ShieldStaff(34,15)
        self.max_level_50[20] = ShieldStaff(28,41)
        self.max_level_50[23] = ShieldStaff(9,14)
        self.max_level_50[25] = ShieldStaff(32,46)
        self.max_level_50[28] = ShieldStaff(41,20)
        self.max_level_50[30] = ShieldStaff(39,50)
        self.max_level_50[32] = ShieldStaff(46,25)
        self.max_level_50[34] = ShieldStaff(15,17)
        self.max_level_50[36] = ShieldStaff(44,48)
        self.max_level_50[39] = ShieldStaff(50,30)
        self.max_level_50[41] = ShieldStaff(20,28)
        self.max_level_50[44] = ShieldStaff(48,36)
        self.max_level_50[46] = ShieldStaff(25,32)
        self.max_level_50[48] = ShieldStaff(36,44)
        self.max_level_50[50] = ShieldStaff(30,39)

        self.max_level_99[7] = ShieldStaff(9,17)
        self.max_level_99[9] = ShieldStaff(17,7)
        self.max_level_99[12] = ShieldStaff(12,12)
        self.max_level_99[15] = ShieldStaff(23,33)
        self.max_level_99[17] = ShieldStaff(7,9)
        self.max_level_99[20] = ShieldStaff(20,20)
        self.max_level_99[23] = ShieldStaff(33,15)
        self.max_level_99[25] = ShieldStaff(28,47)
        self.max_level_99[28] = ShieldStaff(47,25)
        self.max_level_99[31] = ShieldStaff(39,59)
        self.max_level_99[33] = ShieldStaff(15,23)
        self.max_level_99[36] = ShieldStaff(44,73)
        self.max_level_99[39] = ShieldStaff(59,31)
        self.max_level_99[41] = ShieldStaff(53,99)
        self.max_level_99[44] = ShieldStaff(73,36)
        self.max_level_99[47] = ShieldStaff(25,28)
        self.max_level_99[49] = ShieldStaff(65,85)
        self.max_level_99[53] = ShieldStaff(99,41)
        self.max_level_99[59] = ShieldStaff(31,39)
        self.max_level_99[65] = ShieldStaff(85,49)
        self.max_level_99[73] = ShieldStaff(36,44)
        self.max_level_99[85] = ShieldStaff(49,65)
        self.max_level_99[99] = ShieldStaff(41,53)

        self.sword_to_shield_50 = {}
        self.sword_to_staff_50 = {}
        self.sword_to_shield_99 = {}
        self.sword_to_staff_99 = {}
        for i in range(1,100):
            mapping_50 = self.max_level_50[i]
            mapping_99 = self.max_level_99[i]
            if mapping_50 is not None:
                self.sword_to_shield_50[i] = mapping_50.shield
                self.sword_to_staff_50[i] = mapping_50.staff
            else:
                self.sword_to_shield_50[i] = i
                self.sword_to_staff_50[i] = i
            if mapping_99 is not None:
                self.sword_to_shield_99[i] = mapping_99.shield
                self.sword_to_staff_99[i] = mapping_99.staff
            else:
                self.sword_to_shield_99[i] = i
                self.sword_to_staff_99[i] = i


    def get_item_lookup_for_shield(self,max_level,shield_level):
        if max_level==50:
            looking_table = self.sword_to_shield_50
        elif max_level==99:
            looking_table = self.sword_to_shield_99
        else:
            raise ValueError(f"Tried accessing a level progression with max level {max_level}")

        for key,it in looking_table.items():
            if it == shield_level:
                return key

        return None

    def get_item_lookup_for_staff(self,max_level,staff_level):
        if max_level==50:
            looking_table = self.sword_to_staff_50
        elif max_level==99:
            looking_table = self.sword_to_staff_99
        else:
            raise ValueError(f"Tried accessing a level progression with max level {max_level}")

        for key,it in looking_table.items():
            if it == staff_level:
                return key

        return None


        

    def get_shield_level(self,max_level,sword_level):
        if max_level==50:
            return self.sword_to_shield_50[sword_level]
        elif max_level==99:
            return self.sword_to_shield_99[sword_level]
        elif max_level==1:
            return sword_level
        else:
            raise ValueError(f"Tried accessing a level progression with max level {max_level}")

    def get_staff_level(self,max_level,sword_level):
        if max_level==50:
            return self.sword_to_staff_50[sword_level]
        elif max_level==99:
            return self.sword_to_staff_99[sword_level]
        elif max_level==1:
            return sword_level
        else:
            raise ValueError(f"Tried accessing a level progression with max level {max_level}")


        


from dataclasses import dataclass, field

from List.configDict import itemType
from List.inventory.item import InventoryItem


@dataclass(frozen=True)
class Ability(InventoryItem):
    id: int
    name: str
    type: itemType
    sort_index: int


@dataclass(frozen=True)
class DonaldAbility(InventoryItem):
    id: int
    name: str
    type: itemType = field(init=False, default=itemType.DONALD_ABILITY)


@dataclass(frozen=True)
class GoofyAbility(InventoryItem):
    id: int
    name: str
    type: itemType = field(init=False, default=itemType.GOOFY_ABILITY)


Scan = Ability(138, "Scan", itemType.SUPPORT_ABILITY, sort_index=100)
AerialRecovery = Ability(158, "Aerial Recovery", itemType.SUPPORT_ABILITY, sort_index=101)
ComboPlus = Ability(162, "Combo Plus", itemType.SUPPORT_ABILITY, sort_index=103)
AirComboPlus = Ability(163, "Air Combo Plus", itemType.SUPPORT_ABILITY, sort_index=104)
ComboBoost = Ability(390, "Combo Boost", itemType.SUPPORT_ABILITY, sort_index=105)
AirComboBoost = Ability(391, "Air Combo Boost", itemType.SUPPORT_ABILITY, sort_index=106)
ReactionBoost = Ability(392, "Reaction Boost", itemType.SUPPORT_ABILITY, sort_index=107)
FinishingPlus = Ability(393, "Finishing Plus", itemType.SUPPORT_ABILITY, sort_index=108)
NegativeCombo = Ability(394, "Negative Combo", itemType.SUPPORT_ABILITY, sort_index=109)
BerserkCharge = Ability(395, "Berserk Charge", itemType.SUPPORT_ABILITY, sort_index=111)
DamageDrive = Ability(396, "Damage Drive", itemType.SUPPORT_ABILITY, sort_index=112)
DriveBoost = Ability(397, "Drive Boost", itemType.SUPPORT_ABILITY, sort_index=113)
FormBoost = Ability(398, "Form Boost", itemType.SUPPORT_ABILITY, sort_index=114)
SummonBoost = Ability(399, "Summon Boost", itemType.SUPPORT_ABILITY, sort_index=115)
CombinationBoost = Ability(400, "Combination Boost", itemType.SUPPORT_ABILITY, sort_index=116)
ExperienceBoost = Ability(401, "Experience Boost", itemType.SUPPORT_ABILITY, sort_index=117)
LeafBracer = Ability(402, "Leaf Bracer", itemType.SUPPORT_ABILITY, sort_index=118)
MagicLockOn = Ability(403, "Magic Lock-On", itemType.SUPPORT_ABILITY, sort_index=119)
NoExperience = Ability(404, "No Experience", itemType.SUPPORT_ABILITY, sort_index=141)
Draw = Ability(405, "Draw", itemType.SUPPORT_ABILITY, sort_index=121)
Jackpot = Ability(406, "Jackpot", itemType.SUPPORT_ABILITY, sort_index=122)
LuckyLucky = Ability(407, "Lucky Lucky", itemType.SUPPORT_ABILITY, sort_index=123)
FireBoost = Ability(408, "Fire Boost", itemType.SUPPORT_ABILITY, sort_index=125)
BlizzardBoost = Ability(409, "Blizzard Boost", itemType.SUPPORT_ABILITY, sort_index=126)
ThunderBoost = Ability(410, "Thunder Boost", itemType.SUPPORT_ABILITY, sort_index=127)
ItemBoost = Ability(411, "Item Boost", itemType.SUPPORT_ABILITY, sort_index=128)
MpRage = Ability(412, "MP Rage", itemType.SUPPORT_ABILITY, sort_index=129)
MpHaste = Ability(413, "MP Haste", itemType.SUPPORT_ABILITY, sort_index=131)
Defender = Ability(414, "Defender", itemType.SUPPORT_ABILITY, sort_index=134)
SecondChance = Ability(415, "Second Chance", itemType.SUPPORT_ABILITY, sort_index=136)
OnceMore = Ability(416, "Once More", itemType.SUPPORT_ABILITY, sort_index=137)
MpHastera = Ability(421, "MP Hastera", itemType.SUPPORT_ABILITY, sort_index=132)
MpHastega = Ability(422, "MP Hastega", itemType.SUPPORT_ABILITY, sort_index=133)
ComboMaster = Ability(539, "Combo Master", itemType.SUPPORT_ABILITY, sort_index=102)
DriveConverter = Ability(540, "Drive Converter", itemType.SUPPORT_ABILITY, sort_index=124)
LightAndDarkness = Ability(541, "Light & Darkness", itemType.SUPPORT_ABILITY, sort_index=120)
DamageControl = Ability(542, "Damage Control", itemType.SUPPORT_ABILITY, sort_index=135)

Guard = Ability(82, "Guard", itemType.ACTION_ABILITY, sort_index=0)
UpperSlash = Ability(137, "Upper Slash", itemType.ACTION_ABILITY, sort_index=1)
TrinityLimit = Ability(198, "Trinity Limit", itemType.ACTION_ABILITY, sort_index=24)
Slapshot = Ability(262, "Slapshot", itemType.ACTION_ABILITY, sort_index=5)
DodgeSlash = Ability(263, "Dodge Slash", itemType.ACTION_ABILITY, sort_index=6)
SlideDash = Ability(264, "Slide Dash", itemType.ACTION_ABILITY, sort_index=8)
GuardBreak = Ability(265, "Guard Break", itemType.ACTION_ABILITY, sort_index=10)
Explosion = Ability(266, "Explosion", itemType.ACTION_ABILITY, sort_index=11)
FinishingLeap = Ability(267, "Finishing Leap", itemType.ACTION_ABILITY, sort_index=3)
Counterguard = Ability(268, "Counterguard", itemType.ACTION_ABILITY, sort_index=17)
AerialSweep = Ability(269, "Aerial Sweep", itemType.ACTION_ABILITY, sort_index=12)
AerialSpiral = Ability(270, "Aerial Spiral", itemType.ACTION_ABILITY, sort_index=14)
HorizontalSlash = Ability(271, "Horizontal Slash", itemType.ACTION_ABILITY, sort_index=2)
AerialFinish = Ability(272, "Aerial Finish", itemType.ACTION_ABILITY, sort_index=15)
RetaliatingSlash = Ability(273, "Retaliating Slash", itemType.ACTION_ABILITY, sort_index=4)
AutoValor = Ability(385, "Auto Valor", itemType.ACTION_ABILITY, sort_index=18)
AutoWisdom = Ability(386, "Auto Wisdom", itemType.ACTION_ABILITY, sort_index=19)
AutoMaster = Ability(387, "Auto Master", itemType.ACTION_ABILITY, sort_index=21)
AutoFinal = Ability(388, "Auto Final", itemType.ACTION_ABILITY, sort_index=22)
AutoSummon = Ability(389, "Auto Summon", itemType.ACTION_ABILITY, sort_index=23)
FlashStep = Ability(559, "Flash Step", itemType.ACTION_ABILITY, sort_index=7)
AerialDive = Ability(560, "Aerial Dive", itemType.ACTION_ABILITY, sort_index=13)
MagnetBurst = Ability(561, "Magnet Burst", itemType.ACTION_ABILITY, sort_index=16)
VicinityBreak = Ability(562, "Vicinity Break", itemType.ACTION_ABILITY, sort_index=9)
AutoLimitForm = Ability(568, "Auto Limit", itemType.ACTION_ABILITY, sort_index=20)


class DonaldAbilities:
    # These are in a separate namespace because some share IDs with common and/or Goofy abilities
    DonaldFire = DonaldAbility(165, "Donald Fire")
    DonaldBlizzard = DonaldAbility(166, "Donald Blizzard")
    DonaldThunder = DonaldAbility(167, "Donald Thunder")
    DonaldCure = DonaldAbility(168, "Donald Cure")
    Fantasia = DonaldAbility(199, "Fantasia (Comet)")
    FlareForce = DonaldAbility(200, "Flare Force (Duckflare)")
    Draw = DonaldAbility(405, "Draw")
    Jackpot = DonaldAbility(406, "Jackpot")
    LuckyLucky = DonaldAbility(407, "Lucky Lucky")
    FireBoost = DonaldAbility(408, "Fire Boost")
    BlizzardBoost = DonaldAbility(409, "Blizzard Boost")
    ThunderBoost = DonaldAbility(410, "Thunder Boost")
    ItemBoost = DonaldAbility(411, "Item Boost")
    MpRage = DonaldAbility(412, "MP Rage")
    MpHaste = DonaldAbility(413, "MP Haste")
    AutoLimitParty = DonaldAbility(417, "Auto Limit")
    HyperHealing = DonaldAbility(419, "Hyper Healing")
    AutoHealing = DonaldAbility(420, "Auto Healing")
    MpHastera = DonaldAbility(421, "MP Hastera")
    MpHastega = DonaldAbility(422, "MP Hastega")
    DamageControl = DonaldAbility(542, "Damage Control")


class GoofyAbilities:
    # These are in a separate namespace because some share IDs with common and/or Donald abilities
    TornadoFusion = GoofyAbility(201, "Tornado Fusion (Whirligoof)")
    Teamwork = GoofyAbility(202, "Teamwork (Knocksmash)")
    Draw = GoofyAbility(405, "Draw")
    Jackpot = GoofyAbility(406, "Jackpot")
    LuckyLucky = GoofyAbility(407, "Lucky Lucky")
    ItemBoost = GoofyAbility(411, "Item Boost")
    MpRage = GoofyAbility(412, "MP Rage")
    MpHaste = GoofyAbility(413, "MP Haste")
    Defender = GoofyAbility(414, "Defender")
    SecondChance = GoofyAbility(415, "Second Chance")
    OnceMore = GoofyAbility(416, "Once More")
    AutoLimitParty = GoofyAbility(417, "Auto Limit")
    AutoChange = GoofyAbility(418, "Auto Change")
    HyperHealing = GoofyAbility(419, "Hyper Healing")
    AutoHealing = GoofyAbility(420, "Auto Healing")
    MpHastera = GoofyAbility(421, "MP Hastera")
    MpHastega = GoofyAbility(422, "MP Hastega")
    GoofyTornado = GoofyAbility(423, "Goofy Tornado")
    GoofyTurbo = GoofyAbility(425, "Goofy Turbo")
    GoofyBash = GoofyAbility(429, "Goofy Bash")
    DamageControl = GoofyAbility(542, "Damage Control")
    Protect = GoofyAbility(596, "Protect")
    Protera = GoofyAbility(597, "Protera")
    Protega = GoofyAbility(598, "Protega")



import random

hashTextEntries = [
    "item-consumable",
    "item-tent",
    "item-key",
    "ability-unequip",
    "weapon-keyblade",
    "weapon-staff",
    "weapon-shield",
    "armor",
    "magic",
    "material",
    "exclamation-mark",
    "question-mark",
    "accessory",
    "party",
    "ai-mode-frequent",
    "ai-mode-moderate",
    "ai-mode-rare",
    "ai-settings",
    "rank-s",
    "rank-a",
    "rank-b",
    "rank-c",
    "gumi-brush",
    "gumi-blueprint",
    "gumi-ship",
    "gumi-block",
    "gumi-gear"
# These were commented out due to the PC and PS2 versions not having matching icons for them
#    "form",
#    "button-r1",
#    "button-r2",
#    "button-l1",
#    "button-l2",
#    "button-triangle",
#    "button-cross",
#    "button-square",
#    "button-circle",
]


def generate_hash_icons() -> list[str]:
    return random.choices(hashTextEntries, k=7)



from dataclasses import dataclass

from List.configDict import itemType
from List.inventory.item import InventoryItem


@dataclass(frozen=True)
class MagicElement(InventoryItem):
    id: int
    name: str
    type: itemType


Fire = MagicElement(21, "Fire Element", itemType.FIRE)
Blizzard = MagicElement(22, "Blizzard Element", itemType.BLIZZARD)
Thunder = MagicElement(23, "Thunder Element", itemType.THUNDER)
Cure = MagicElement(24, "Cure Element", itemType.CURE)
Magnet = MagicElement(87, "Magnet Element", itemType.MAGNET)
Reflect = MagicElement(88, "Reflect Element", itemType.REFLECT)



from typing import Optional

from Class.itemClass import KH2Item
from List.ItemList import Items
from List.ObjectiveList import KH2Objective
from List.configDict import locationType
from List.inventory import misc
from Module.newRandomize import ItemAssignment, SynthesisRecipe, WeaponStats
from Module.weighting import LocationWeights


def item_spoiler_dictionary(
        item_assignments: list[ItemAssignment],
        starting_inventory_ids: Optional[list[int]] = None,
        shop_items: Optional[list[KH2Item]] = None,
        weights: LocationWeights = None,
        location_spheres=None
) -> dict[locationType, list[tuple[str, KH2Item]]]:
    # (depth for sort purposes, location id for sort purposes, sphere, spoiler log string, item)
    out_dict: dict[locationType, list[tuple[int, int, str, str, KH2Item]]] = {}

    item_lookup = Items.sora_lookup_table()
    if starting_inventory_ids is not None:
        item_spoiler = []
        for index, starting_item_id in enumerate(starting_inventory_ids):
            item = item_lookup.get(starting_item_id)
            if item is not None:
                item_spoiler.append((0, index, 0, "Sora Starting Item", item))
            else:
                print(f"Could not find item with id {starting_item_id} to add to the spoiler log")
        if len(item_spoiler) > 0:
            out_dict[locationType.Free] = item_spoiler

    if shop_items is not None:
        shop_spoiler = []
        for shop_item in shop_items:
            shop_spoiler.append((0, 0, 0, "Moogle Shop", shop_item))
        if len(shop_spoiler) > 0:
            out_dict[locationType.SHOP] = shop_spoiler

    for assignment in item_assignments:
        location = assignment.location
        loc_id = location.LocationId

        sort_depth = 0
        if weights and location in weights.location_depths:
            depth = weights.get_depth(location)
            added_string = f" <{depth}>"
            if isinstance(depth, int):
                sort_depth = depth
            elif isinstance(depth, tuple):
                sort_depth = depth[0]
        else:
            added_string = ""

        sphere = "N/A"
        prepend_string = ""
        if location_spheres:
            if location not in location_spheres:
                prepend_string = "Unreachable "
            elif location in location_spheres:
                prepend_string = ""
                sphere = str(location_spheres[location])

        spoiler_string = prepend_string + location.Description + added_string

        if len(location.LocationTypes) > 0:
            location_type = location.LocationTypes[0]
            if location_type not in out_dict:
                out_dict[location_type] = []
            if assignment.item.item != misc.NullItem:
                out_dict[location_type].append((sort_depth, loc_id, sphere, spoiler_string, assignment.item))
            if assignment.item2 is not None:
                out_dict[location_type].append((sort_depth, loc_id, sphere, spoiler_string, assignment.item2))

    result_dict: dict[locationType, list[tuple[int, str, KH2Item]]] = {}
    for location_type, entries in out_dict.items():
        # Sort first by the location ID (this will be the tiebreaker)
        entries.sort(key=lambda entry: entry[1])
        # Now sort by the depth
        entries.sort(key=lambda entry: entry[0])
        result_dict[location_type] = [(entry[2], entry[3], entry[4]) for entry in entries]

    return result_dict


def levelStatsDictionary(level_stats):
    outDict = {}
    for lvl in level_stats:
        desc = lvl.location.Description
        split_desc=desc.split(' ')
        if len(split_desc) > 2:
            desc = split_desc[0] + " " +split_desc[2]
        outDict[desc] = {"experience":lvl.experience,"strength":lvl.strength, "magic": lvl.magic, "defense": lvl.defense, "ap":lvl.ap }
    return outDict


def synth_recipe_dictionary(assignments: list[ItemAssignment], recipes: list[SynthesisRecipe]) -> list[dict[str, str]]:
    result: list[dict[str, str]] = []

    sorted_recipes = [r for r in recipes]
    sorted_recipes.sort(key=lambda r: r.location.LocationId)

    for recipe in sorted_recipes:
        location = recipe.location
        synth_assignment = next(a for a in assignments if a.location == location)
        requirement_strings = []
        for requirement in recipe.requirements:
            requirement_strings.append(" x".join([requirement.synth_item.Name, f"{requirement.amount}"]))
        result.append({
            "location": location.Description,
            "item": synth_assignment.item.Name,
            "requirements": ", ".join(requirement_strings)
        })

    return result


def weapon_stats_dictionary(
        sora_assignments: list[ItemAssignment],
        donald_assignments: list[ItemAssignment],
        goofy_assignments: list[ItemAssignment],
        weapons: list[WeaponStats]
) -> list[dict[str, str]]:
    all_assignments = sora_assignments + donald_assignments + goofy_assignments

    result: list[dict[str, str]] = []
    for weapon in weapons:
        location = weapon.location
        assignment = next(a for a in all_assignments if a.location == location)
        result.append({
            "name": location.Description.partition(" (Slot)")[0],
            "strength": weapon.strength,
            "magic": weapon.magic,
            "item": assignment.item.Name
        })
    return result


def objectives_dictionary(objectives: list[KH2Objective]) -> list[dict[str, str]]:
    result: list[dict[str, str]] = []
    for objective in objectives:
        result.append({
            "name": objective.Name,
            "type": objective.Type.value,
            "location": objective.Location.value,
            "difficulty": objective.Difficulty.value,
        })
    return result
